# A module for feature/peak detection

# Import modules
import numpy as np
from scipy.signal import find_peaks
from scipy.ndimage import gaussian_filter1d
from copy import deepcopy
import bisect

from masscube.feature_evaluation import calculate_noise_level, calculate_gaussian_similarity, calculate_asymmetry_factor


def find_rois(d):
    """
    A function to find the region of interest (ROI) in the MS data.

    Parameters
    ----------------------------------------------------------
    d: MSData object
        An MSData object that contains the MS data.
    """

    # A list to store the rois in progress
    rois = []
    # A list for the finally detected ROIs
    final_rois = []

    # Initiate a set of rois using the first MS1 scan
    fs = d.scans[d.ms1_idx[0]]  # The first scan

    for i in range(len(fs.int_seq)):
        roi = Roi(scan_idx=d.ms1_idx[0], rt=fs.rt, mz=fs.mz_seq[i], intensity=fs.int_seq[i])
        rois.append(roi)

    last_ms1_idx = d.ms1_idx[0]
    last_rt = fs.rt

    # Loop over all MS1 scans
    for ms1_idx in d.ms1_idx[1:]:

        s = d.scans[ms1_idx]  # The current MS1 scan

        if not s:
            continue

        if len(s.mz_seq) == 0:
            continue

        visited_idx = []  # A list to store the visited indices of ions in the current MS1 scan
        visited_rois_idx = []  # A list to store the visited indices of rois

        # Loop over all current rois
        for i, roi in enumerate(rois):

            min_idx = find_closest_index_ordered(s.mz_seq, roi.mz_seq[-1], d.params.mz_tol_ms1)
            if min_idx is not None and min_idx not in visited_idx:
                roi.extend_roi(scan_idx=ms1_idx, rt=s.rt, mz=s.mz_seq[min_idx], intensity=s.int_seq[min_idx])
                roi.gap_counter = 0
                visited_idx.append(min_idx)
                visited_rois_idx.append(i)

        to_be_moved = []

        # Plus one to the gap counter of the rois that are not visited
        for i in range(len(rois)):
            if i not in visited_rois_idx:
                rois[i].extend_roi(scan_idx=ms1_idx, rt=s.rt, mz=rois[i].mz_seq[-1], intensity=0)
                rois[i].gap_counter = rois[i].gap_counter + 1
                if rois[i].gap_counter > d.params.roi_gap:
                    to_be_moved.append(i)

        # Move the rois that have not been visited for a long time to final_rois
        for i in to_be_moved[::-1]:
            final_rois.append(rois.pop(i))

        # Create new rois for the rest
        for i in range(len(s.int_seq)):
            if i not in visited_idx:
                # Add a zero before the new roi
                roi = Roi(scan_idx=last_ms1_idx, rt=last_rt, mz=s.mz_seq[i], intensity=0)
                roi.extend_roi(scan_idx=ms1_idx, rt=s.rt, mz=s.mz_seq[i], intensity=s.int_seq[i])
                rois.append(roi)

        rois.sort(key=lambda x: x.int_seq[-1], reverse=True)
        last_ms1_idx = ms1_idx
        last_rt = s.rt

    # Move all rois to final_rois
    for roi in rois:
        final_rois.append(roi)

    return final_rois


def cut_roi(r, int_tol=1000, sigma=1.2, prominence_ratio=0.1, distance=5):
    """
    Function to cut an ROI by providing the start and end positions.

    Parameters
    ----------
    r: Roi object
        An ROI object to be cut.
    int_tol: int
        The intensity tolerance for cutting.
    sigma: float
        The sigma value for Gaussian filter.
    prominence_ratio: float
        The prominence ratio for finding peaks. prom = np.max(y)*prominence_ratio

    Returns
    -------
    cut_rois: list
        A list of cut ROIs.
    """

    r.int_seq = np.array(r.int_seq)
    r.noise_level = calculate_noise_level(r.int_seq)

    if r.peak_height < 3 * int_tol or r.noise_level > 0.6:
        return [r]
    if r.noise_level < 0.1:
        prominence_ratio = 0.5 * prominence_ratio
    if len(r.int_seq) < 10:
        sigma = 0.5 * sigma
        distance = 3
    ss = gaussian_filter1d(r.int_seq, sigma=sigma)
    peaks, _ = find_peaks(ss, prominence=np.max(ss) * prominence_ratio, distance=distance)

    peaks = peaks[r.int_seq[peaks] > 3 * int_tol]

    if len(peaks) < 2:
        return [r]

    if len(peaks) > 4:
        peaks = np.sort(peaks[np.argsort(r.int_seq[peaks])[-4:]])

    positions = [0]
    for i in range(len(peaks) - 1):
        lowest_int = 1e10
        for j in range(peaks[i], peaks[i + 1]):
            if r.int_seq[j] < lowest_int:
                lowest_int = r.int_seq[j]
                lowest_int_idx = j
        positions.append(lowest_int_idx)
    positions.append(len(r.int_seq) - 1)

    cut_rois = []
    for i in range(len(positions) - 1):
        tmp = deepcopy(r)
        tmp.subset_roi(start=positions[i], end=positions[i + 1] + 1)
        tmp.cut = True
        cut_rois.append(tmp)

    return cut_rois


class Roi:
    """
    A class to store a region of interest (ROI).
    """

    def __init__(self, scan_idx, rt, mz, intensity):
        """
        Function to initiate an ROI by providing the scan indices,
        retention time, m/z and intensity.

        Parameters
        ----------------------------------------------------------
        scan_idx: int
            Scan index of the first ion in the ROI
        rt: float
            Retention time of the first ion in the ROI
        mz: float
            m/z value of the first ion in the ROI
        intensity: int
            Intensity of the first ion in the ROI
        """

        self.id = None
        self.scan_idx_seq = [scan_idx]

        self.rt_seq = [rt]
        self.mz_seq = [mz]
        self.int_seq = [intensity]
        self.ms2_seq = []

        # Count the gaps in the ROI's tail
        self.gap_counter = 0

        # Create attributes for the summarized values of the ROI
        self.mz = mz
        self.rt = np.nan
        self.scan_number = 0
        self.peak_area = np.nan
        self.peak_height = np.nan
        self.best_ms2 = None
        self.length = 0
        self.merged = False
        self.gaussian_similarity = 0.0
        self.noise_level = 0.0
        self.asymmetry_factor = 0.0
        self.cut = False

        # Isotopes
        self.charge_state = 1
        self.is_isotope = False
        self.isotope_mz_seq = []
        self.isotope_int_seq = []
        self.isotope_id_seq = []

        # In-source fragments
        self.is_in_source_fragment = False
        self.isf_child_roi_id = []
        self.isf_parent_roi_id = None

        # Adducts
        self.adduct_type = None
        self.adduct_parent_roi_id = None
        self.adduct_child_roi_id = []

        # Annotation
        self.annotation = None
        self.formula = None
        self.similarity = None
        self.matched_peak_number = None
        self.smiles = None
        self.inchikey = None

    def extend_roi(self, scan_idx, rt, mz, intensity):
        """
        Function to extend an ROI by providing the scan indices,
        retention time, m/z and intensity.

        Parameters
        ----------------------------------------------------------
        scan_idx: int
            Scan index of the ion to be added to the ROI
        rt: float
            Retention time of the ion to be added to the ROI
        mz: float
            m/z value of the ion to be added to the ROI
        intensity: int
            Intensity of the ion to be added to the ROI
        """

        # Extend the ROI
        self.scan_idx_seq.append(scan_idx)
        self.rt_seq.append(rt)
        self.mz_seq.append(mz)
        self.int_seq.append(intensity)

    def show_roi_info(self, show_annotation=False):
        """
        Function to print the information of the ROI.
        """

        print(f"ROI: {self.mz:.4f} m/z, {self.rt:.2f} min, {self.peak_area:.2f} area, {self.peak_height:.2f} height")
        print(f"ROI start time: {self.rt_seq[0]:.2f} min, ROI end time: {self.rt_seq[-1]:.2f} min")

        if show_annotation:
            # show isotopes, in-source fragments and adducts
            print("Isotope information:")
            print(f"Isotope charge state: {self.charge_state}")
            print(f"Isotope state: {self.isotope_state}")
            print(f"Isotope m/z: {self.isotope_mz_seq}")
            print(f"Isotope intensity: {self.isotope_int_seq}")

            print("In-source fragment information:")
            print(f"In-source fragment: {self.is_in_source_fragment}")
            print(f"Isf child roi id: {self.isf_child_roi_id}")
            print(f"Isf parent roi id: {self.isf_parent_roi_id}")

            print("Adduct information:")
            print(f"Adduct type: {self.adduct_type}")
            print(f"Adduct parent roi id: {self.adduct_parent_roi_id}")
            print(f"Adduct child roi id: {self.adduct_child_roi_id}")

    def get_mz_error(self):
        """
        Function to calculate the m/z error (maximum - minimum) of the ROI.
        """

        return np.max(self.mz_seq) - np.min(self.mz_seq)

    def find_rt_ph_pa(self):
        """
        Function to find the peak area of the ROI using trapzoidal rule.

        """

        idx = np.argmax(self.int_seq)
        self.mz = self.mz_seq[idx]
        self.rt = self.rt_seq[idx]
        self.peak_height = self.int_seq[idx]
        self.peak_area = int(np.trapz(y=self.int_seq, x=self.rt_seq) * 60)

    def find_top_average(self, num=3):
        """
        Function to find the peak height of the ROI by averaging
        the heighest three intensities.
        """

        d = np.sort(self.int_seq)[-num:]
        # calculate mean of non-zero values
        d = d[d != 0]
        self.top_average = np.mean(d, dtype=np.int64)

    def sum_roi(self, cal_g_score=True, cal_a_score=True):
        """
        Function to summarize the ROI to generate attributes.
        """
        self.int_seq = np.array(self.int_seq)
        end_idx = len(self.int_seq) - 1

        while self.int_seq[end_idx] == 0 and self.int_seq[end_idx - 1] == 0:
            end_idx -= 1

        end_idx += 1

        # keep one zero in the end of ROI
        self.mz_seq = self.mz_seq[:end_idx]
        self.int_seq = self.int_seq[:end_idx]
        self.rt_seq = self.rt_seq[:end_idx]
        self.scan_idx_seq = self.scan_idx_seq[:end_idx]

        self.find_rt_ph_pa()

        self.noise_level = calculate_noise_level(self.int_seq)

        if cal_g_score:
            self.gaussian_similarity = calculate_gaussian_similarity(self.rt_seq, self.int_seq)
        if cal_a_score:
            self.asymmetry_factor = calculate_asymmetry_factor(self.int_seq)

        self.length = np.sum(self.int_seq > 0)

    def subset_roi(self, start, end):
        """
        Function to subset the ROI by providing the positions.

        Parameters
        ----------------------------------------------------------
        start: int
            The start position of the ROI to be subsetted.
        end: int
            The end position of the ROI to be subsetted.
        """

        self.scan_idx_seq = self.scan_idx_seq[start:end]
        self.rt_seq = self.rt_seq[start:end]
        self.mz_seq = self.mz_seq[start:end]
        self.int_seq = self.int_seq[start:end]

        ms2_seq = []

        for ms2 in self.ms2_seq:
            if ms2.scan > self.scan_idx_seq[0] and ms2.scan < self.scan_idx_seq[-1]:
                ms2_seq.append(ms2)

        self.ms2_seq = ms2_seq


def find_closest_index_ordered(array, target, tol=0.01):
    """
    Function to find the index of the closest value in an ordered array.
    """

    idx = bisect.bisect_left(array, target)

    if idx == 0:
        if array[idx] - target < tol:
            return 0
        else:
            return None
    if idx == len(array):
        if target - array[idx - 1] < tol:
            return len(array) - 1
        else:
            return None

    before = array[idx - 1]
    after = array[idx]

    if after - target < target - before and after - target < tol:
        return idx
    elif after - target > target - before and target - before < tol:
        return idx - 1
    else:
        return None


import pandas as pd

import matplotlib.pyplot as plt
import numpy as np

from json import loads as loads
from requests import get


def create_mirror_plot(mz_arr, int_arr, mz_arr_ref, int_arr_ref,
                       ms2_tol=0.05, peak_int_power=0.5,
                       intensity_threshold_label=0, label_decimals=2,
                       max_x=None,
                       matched_peak_color='red',
                       up_spec_peak_color='0.6', down_spec_peak_color='0.6',
                       save=False,
                       output_name='demo.svg'):

    plt.figure(figsize=(4.5, 1.7))

    int_arr = np.power(int_arr, peak_int_power)
    int_arr_ref = np.power(int_arr_ref, peak_int_power)

    # Normalize intensities
    max_intensity = max(int_arr[mz_arr < max_x])
    int_arr = int_arr / max_intensity * 100

    max_intensity_ref = max(int_arr_ref[mz_arr_ref < max_x])
    int_arr_ref = int_arr_ref / max_intensity_ref * 100

    # Find matched peaks
    matched_indices = []
    matched_indices_ref = []
    for i, mz in enumerate(mz_arr):
        matches = np.where(np.abs(mz_arr_ref - mz) <= ms2_tol)[0]
        if matches.size > 0:
            print(f'{mz:.4f} matched to {mz_arr_ref[matches]}')
            matched_indices.extend([i] * len(matches))
            matched_indices_ref.extend(matches)

    # Set axis limits and labels
    if max_x is None:
        max_x = max(np.max(mz_arr), np.max(mz_arr_ref)) * 1.05  # Add 5% margin

    # Plot unmatched peaks
    plt.vlines(mz_arr, 0, int_arr, color=up_spec_peak_color, linewidth=1.5)
    plt.vlines(mz_arr_ref, 0, -int_arr_ref, color=down_spec_peak_color, linewidth=1.5)

    # Plot matched peaks
    plt.vlines(mz_arr[matched_indices], 0, int_arr[matched_indices], color=matched_peak_color, linewidth=1.5)
    plt.vlines(mz_arr_ref[matched_indices_ref], 0, -int_arr_ref[matched_indices_ref], color=matched_peak_color,
               linewidth=1.5)

    plt.xlim(0, max_x)
    plt.ylim(-110, 110)

    # Add labels for peaks above threshold
    for mz, intensity in zip(mz_arr, int_arr):
        if intensity >= intensity_threshold_label and mz < max_x:
            plt.text(mz, intensity + 0.1, f'{mz:.{label_decimals}f}', ha='center', va='bottom', fontsize=8, fontname='Arial')

    for mz, intensity in zip(mz_arr_ref, int_arr_ref):
        if intensity >= intensity_threshold_label and mz < max_x:
            plt.text(mz, -intensity - 2, f'{mz:.{label_decimals}f}', ha='center', va='top', fontsize=8, fontname='Arial')

    plt.xlabel(r'$\mathit{m/z}$', fontname='Arial', fontsize=12, labelpad=2, color='0.2')
    plt.ylabel('Intensity (%)', fontname='Arial', fontsize=12, labelpad=2, color='0.2')

    plt.gca().xaxis.set_visible(False)

    # Remove top and right spines
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['bottom'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    # Set the color of the frame (axes spines)
    for spine in plt.gca().spines.values():
        spine.set_color('0.5')

    # Remove grid
    plt.grid(False)

    # Add line at y=0
    plt.axhline(y=0, color='0.5', linewidth=1)

    # Use Helvetica font for tick labels
    # plt.xticks(fontname='Arial', fontsize=10)
    plt.yticks(fontname='Arial', fontsize=10)

    # Adjust tick parameters
    # plt.tick_params(axis='x', pad=4, length=2, color='0.5')
    plt.tick_params(axis='y', pad=2.5, length=2, color='0.5')

    # Set the color of tick labels to 0.2
    plt.gca().tick_params(axis='both', colors='0.2')

    # Modify y-axis ticks and labels
    yticks = plt.gca().get_yticks()
    yticks = [y for y in yticks if abs(y) <= 100]
    plt.gca().set_yticks(yticks)
    plt.gca().set_yticklabels([str(abs(int(y))) for y in yticks])

    # Save the plot
    plt.tight_layout()
    if save:
        plt.savefig(output_name, transparent=True)
    plt.show()
    plt.close()


def demo():
    df1 = pd.read_csv(
        '/data/nist/fullscan_0ev/aligned_feature_table.tsv',
        sep='\t', low_memory=False)

    df1 = df1[df1['MS1_annotation'] == 'Phenylalanine_C3:0'].iloc[0]

    ms2_1 = df1['pseudo_ms2']
    mz_arr1 = []
    int_arr1 = []
    for p in ms2_1.split(';'):
        p = p.strip()
        if not p:
            continue
        mz, inten = p.split(' ')
        mz_arr1.append(float(mz))
        int_arr1.append(float(inten))

    df2 = pd.read_csv(
        '/data/PR000639/output/hilic_pos_aligned_feature_table.tsv',
        sep='\t', low_memory=False)

    df2 = df2[df2['ID'] == 6696].iloc[0]

    ms2_2 = df2['pseudo_ms2']
    mz_arr2 = []
    int_arr2 = []
    for p in ms2_2.split(';'):
        p = p.strip()
        if not p:
            continue
        mz, inten = p.split(' ')
        mz_arr2.append(float(mz))
        int_arr2.append(float(inten))

    print('precursor m/z: ', df2['m/z'])

    create_mirror_plot(np.array(mz_arr1), np.array(int_arr1),
                       np.array(mz_arr2), np.array(int_arr2),
                       ms2_tol=0.05, peak_int_power=0.5,
                       intensity_threshold_label=50,
                       label_decimals=2,
                       max_x=float(df2['m/z']) * 1.05,
                       matched_peak_color='#F08976',
                       up_spec_peak_color='0.9', down_spec_peak_color='0.9',
                       save=True,
                       output_name='demo.svg')


if __name__ == '__main__':

    demo()




import numpy as np
import matplotlib.pyplot as plt
from scipy import sparse


def correlation_summary(cor_matrix, thresholds=[0.90, 0.80, 0.70]):
    """
    Summarize correlation scores and create a histogram plot for a csr_matrix.

    :param cor_matrix: scipy.sparse.csr_matrix of correlation scores
    :param thresholds: List of correlation thresholds to report
    """
    if not sparse.issparse(cor_matrix) or not isinstance(cor_matrix, sparse.csr_matrix):
        raise ValueError("Input must be a scipy.sparse.csr_matrix")

    # Get upper triangle (excluding diagonal)
    upper_tri = sparse.triu(cor_matrix, k=1)

    # Convert to array (this step might be memory-intensive for very large matrices)
    upper_tri_array = upper_tri.data

    # Print summary statistics
    print(f"Total number of correlations: {len(upper_tri_array)}")
    print(f"Mean correlation: {np.mean(upper_tri_array):.4f}")
    print(f"Median correlation: {np.median(upper_tri_array):.4f}")

    for threshold in thresholds:
        count = np.sum(upper_tri_array >= threshold)
        percentage = (count / len(upper_tri_array)) * 100
        print(f"Correlations >= {threshold}: {count} ({percentage:.2f}%)")

    # Create histogram
    plt.figure(figsize=(10, 6))
    plt.hist(upper_tri_array, bins=np.arange(-1.01, 1.011, 0.02), color='#56648a', edgecolor='none')
    plt.title('Histogram of correlation scores')
    plt.xlabel('Correlation score')
    plt.ylabel('Frequency')

    # Add vertical lines for thresholds with different colors
    colors = ['red', 'green', 'blue']  # Standard colors for threshold lines
    for threshold, color in zip(thresholds, colors):
        plt.axvline(x=threshold, color=color, linestyle='--', label=f'Threshold {threshold}')

    plt.legend(loc='upper right')
    plt.tight_layout()

    plt.savefig('correlation_histogram.svg', transparent=True)
    plt.show()


if __name__ == '__main__':
    ################
    ####### careful, in the file folder is filtered mz correlation matrix
    ################ have to recalculate the mz correlation matrix
    file_path = '/Users/shipei/Documents/projects/ms1_id/imaging/MTBLS313/Brain01_Bregma1-42_01_centroid/intensity_matrix.npy'

    intensity_matrix = np.load(file_path)

    from src.ms1_id.msi.calculate_mz_cor_parallel import calc_all_mz_correlations

    mz_cor_matrix = calc_all_mz_correlations(intensity_matrix, min_overlap=10,
                                             min_cor=-1.01,
                                             n_processes=10, save=True,
                                             save_dir='data')

    # mz_cor_matrix = sparse.load_npz('./data/mz_correlation_matrix.npz')

    correlation_summary(mz_cor_matrix)



"""
create a workflow for MS1_ID using masscube backend.
for multiple files.
"""

import multiprocessing
import os

from masscube.alignment import feature_alignment, gap_filling
from masscube.feature_grouping import annotate_isotope
from masscube.feature_table_utils import convert_features_to_df
from masscube.normalization import sample_normalization
from masscube.params import Params
# from masscube.raw_data_utils import MSData
from ms1_id.lcms.masscube_raw_data_utils import MSData

from ms1_id.lcms.annotate_adduct import annotate_adduct
from ms1_id.lcms.annotate_ms2 import feature_annotation
from ms1_id.lcms.calculate_ppc import calc_all_ppc
from ms1_id.lcms.export import write_feature_table
from ms1_id.lcms.group_ppc import generate_pseudo_ms2, retrieve_pseudo_ms2_spectra
from ms1_id.lcms.reverse_matching import ms1_id_annotation
from ms1_id.lcms.utils import find_ms_info

# default parameters
orbitrap_mass_detect_int_tol = 20000
tof_mass_detect_int_tol = 500


def main_workflow(project_path=None, ms1id_library_path=None, ms2id_library_path=None,
                  sample_dir='data', parallel=True,
                  ms1_id=True, ms2_id=False,
                  cpu_ratio=0.9,
                  run_rt_correction=True, run_normalization=False,
                  ms1_tol=0.01, ms2_tol=0.015, mass_detect_int_tol=None,
                  align_mz_tol=0.01, align_rt_tol=0.2, alignment_drop_by_fill_pct_ratio=0.1,
                  peak_cor_rt_tol=0.015,
                  min_ppc=0.6, roi_min_length=3,
                  library_search_mztol=0.05,
                  ms1id_score_cutoff=0.8, ms1id_min_matched_peak=6, ms1id_min_spec_usage=0.05,
                  ms1id_max_prec_rel_int_in_other_ms2=0.01,
                  ms2id_score_cutoff=0.8, ms2id_min_matched_peak=6):
    """
    Main workflow for MS1_ID.
    :param project_path: path to the project directory
    :param ms1id_library_path: path to the MS1 ID library
    :param ms2id_library_path: path to the MS2 ID library
    :param parallel: whether to run in parallel
    :param ms1_id: whether to perform MS1 ID
    :param ms2_id: whether to perform MS2 ID
    :param cpu_ratio: CPU ratio for multiprocessing
    :param sample_dir: directory where the raw files are stored
    :param run_rt_correction: whether to perform RT correction
    :param run_normalization: whether to perform normalization
    :param ms1_tol: m/z tolerance for MS1
    :param ms2_tol: m/z tolerance for MS2
    :param mass_detect_int_tol: intensity tolerance for mass detection
    :param align_mz_tol: alignment m/z tolerance
    :param align_rt_tol: alignment RT tolerance
    :param alignment_drop_by_fill_pct_ratio: alignment drop by fill percentage ratio
    :param peak_cor_rt_tol: peak correlation RT tolerance
    :param min_ppc: minimum peak-peak correlation threshold
    :param roi_min_length: minimum ROI length for feature grouping
    :param ms1id_score_cutoff: ms1 ID score cutoff
    :param ms1id_min_matched_peak: ms1 ID min matched peak
    :param ms1id_max_prec_rel_int_in_other_ms2: ms1 ID, max precursor relative intensity in other MS2
    :param ms2id_score_cutoff: ms2 ID score cutoff
    :param ms2id_min_matched_peak: ms2 ID min matched peak
    :return:
    """

    # init a new config object
    config = init_config(project_path, ms1id_library_path, ms2id_library_path,
                         sample_dir,
                         ms1_id=ms1_id, ms2_id=ms2_id,
                         run_rt_correction=run_rt_correction, run_normalization=run_normalization,
                         mz_tol_ms1=ms1_tol, mz_tol_ms2=ms2_tol, mass_detect_int_tol=mass_detect_int_tol,
                         align_mz_tol=align_mz_tol, align_rt_tol=align_rt_tol,
                         alignment_drop_by_fill_pct_ratio=alignment_drop_by_fill_pct_ratio,
                         peak_cor_rt_tol=peak_cor_rt_tol,
                         min_ppc=min_ppc, roi_min_length=roi_min_length,
                         library_search_mztol=library_search_mztol,
                         ms1id_score_cutoff=ms1id_score_cutoff, ms1id_min_matched_peak=ms1id_min_matched_peak,
                         ms1id_min_spec_usage=ms1id_min_spec_usage,
                         ms1id_max_prec_rel_int_in_other_ms2=ms1id_max_prec_rel_int_in_other_ms2,
                         ms2id_score_cutoff=ms2id_score_cutoff, ms2id_min_matched_peak=ms2id_min_matched_peak)

    raw_file_names = os.listdir(config.sample_dir)
    raw_file_names = [f for f in raw_file_names if f.lower().endswith(".mzml") or f.lower().endswith(".mzxml")]
    raw_file_names = [f for f in raw_file_names if not f.startswith(".")]  # for Mac OS
    total_file_num = len(raw_file_names)

    # skip the files that have been processed
    txt_files = os.listdir(config.single_file_dir)
    txt_files = [f.split(".")[0] for f in txt_files if f.lower().endswith(".txt") and not f.startswith(".")]
    raw_file_names = [f for f in raw_file_names if f.split(".")[0] not in txt_files]
    raw_file_names = [os.path.join(config.sample_dir, f) for f in raw_file_names]

    print("\t{} raw file are found, {} files to be processed.".format(total_file_num, len(raw_file_names)))

    if len(raw_file_names) > 0:
        # process files by multiprocessing
        print("Processing individual files for feature detection, evaluation, and grouping...")

        if parallel:
            workers = min(int(multiprocessing.cpu_count() * cpu_ratio), len(raw_file_names))
            print("\t {} CPU cores in total, {} cores used.".format(multiprocessing.cpu_count(), workers))

            print(f"Processing all {len(raw_file_names)} files")
            p = multiprocessing.Pool(workers)
            p.starmap(feature_detection, [(f, config) for f in raw_file_names])
            p.close()
            p.join()
        else:
            for i, file_name in enumerate(raw_file_names):
                feature_detection(file_name, params=config)

    pseudo_ms2_spectra = []
    if ms1_id:
        # get all pre-saved pseudo ms2 spectra
        pseudo_ms2_spectra = retrieve_pseudo_ms2_spectra(config)

    # feature alignment
    print("Aligning features...")
    features = feature_alignment(config.single_file_dir, config,
                                 drop_by_fill_pct_ratio=alignment_drop_by_fill_pct_ratio)

    # gap filling
    print("Filling gaps...")
    features = gap_filling(features, config)

    # annotation (using MS2 library)
    if ms2_id and config.ms2id_library_path is not None:
        print("Annotating MS2...")
        features = feature_annotation(features, config)

    feature_table = convert_features_to_df(features, config.sample_names)

    # normalization
    if config.run_normalization:
        print("Running normalization...")
        feature_table = sample_normalization(feature_table, config.individual_sample_groups,
                                             config.normalization_method)

    # output feature table
    output_path = os.path.join(config.project_dir, "aligned_feature_table.tsv")
    write_feature_table(feature_table, pseudo_ms2_spectra, config, output_path)

    print("The workflow is completed.")


def init_config(path=None,
                ms1id_library_path=None, ms2id_library_path=None,
                sample_dir='data',
                ms1_id=True, ms2_id=False,
                run_rt_correction=True, run_normalization=False,
                mz_tol_ms1=0.01, mz_tol_ms2=0.015, mass_detect_int_tol=None,
                align_mz_tol=0.01, align_rt_tol=0.2, alignment_drop_by_fill_pct_ratio=0.1,
                peak_cor_rt_tol=0.015,
                min_ppc=0.6, roi_min_length=3,
                library_search_mztol=0.05,
                ms1id_score_cutoff=0.8, ms1id_min_matched_peak=6, ms1id_min_spec_usage=0.05,
                ms1id_max_prec_rel_int_in_other_ms2=0.01,
                ms2id_score_cutoff=0.8, ms2id_min_matched_peak=6
                ):
    # init
    config = Params()

    config.ms1_id = ms1_id
    config.ms2_id = ms2_id

    config.peak_cor_rt_tol = peak_cor_rt_tol
    config.min_ppc = min_ppc
    config.roi_min_length = roi_min_length
    config.ms1id_score_cutoff = ms1id_score_cutoff
    config.ms1id_min_matched_peak = ms1id_min_matched_peak
    config.ms1id_min_spec_usage = ms1id_min_spec_usage
    config.ms1id_max_prec_rel_int_in_other_ms2 = ms1id_max_prec_rel_int_in_other_ms2
    config.ms2id_score_cutoff = ms2id_score_cutoff
    config.ms2id_min_matched_peak = ms2id_min_matched_peak

    # obtain the working directory
    config.project_dir = path
    print("Project directory: " + config.project_dir)

    config.sample_dir = os.path.join(config.project_dir, sample_dir)
    print("Sample directory: " + config.sample_dir)

    # check if the required files are prepared
    if not os.path.exists(config.sample_dir) or len(os.listdir(config.sample_dir)) == 0:
        raise ValueError("No raw MS data is found in the sample directory.")

    config.single_file_dir = os.path.join(config.project_dir, "single_files")
    os.makedirs(config.single_file_dir, exist_ok=True)

    # determine the type of MS and ion mode
    file_names = os.listdir(config.sample_dir)
    file_names = [f for f in file_names if f.lower().endswith(".mzml") or f.lower().endswith(".mzxml")]
    file_names = [f for f in file_names if not f.startswith(".")]  # for Mac OS
    file_name = os.path.join(config.sample_dir, file_names[0])
    ms_type, ion_mode, _ = find_ms_info(file_name)

    if mass_detect_int_tol is not None:
        config.int_tol = mass_detect_int_tol
    else:
        if ms_type == "orbitrap":
            config.int_tol = orbitrap_mass_detect_int_tol
        elif ms_type == "tof":
            config.int_tol = tof_mass_detect_int_tol
        else:
            config.int_tol = 20000

    ##########################
    # The project
    # config.project_dir = None  # Project directory, character string
    config.sample_names = [f.split(".")[0] for f in os.listdir(config.sample_dir)
                           if not f.startswith(".")]
    config.sample_groups = ["sample"] * len(config.sample_names)
    config.individual_sample_groups = ["sample"] * len(config.sample_names)
    config.sample_group_num = 1

    # config.sample_dir = None  # Directory for the sample information, character string
    # config.single_file_dir = None  # Directory for the single file output, character string
    config.annotation_dir = None  # Directory for the annotation output, character string
    config.chromatogram_dir = None  # Directory for the chromatogram output, character string
    config.network_dir = None  # Directory for the network output, character string
    config.statistics_dir = None  # Directory for the statistical analysis output, character string

    # MS data acquisition
    config.rt_range = [0.0, 1000.0]  # RT range in minutes, list of two floats
    config.ion_mode = ion_mode  # Ionization mode, "positive" or "negative", character string

    # Feature detection
    config.mz_tol_ms1 = mz_tol_ms1  # m/z tolerance for MS1, default is 0.01
    config.mz_tol_ms2 = mz_tol_ms2  # m/z tolerance for MS2, default is 0.015
    # config.int_tol = 30000  # Intensity tolerance, default is 30000 for Orbitrap and 1000 for other instruments
    config.roi_gap = 30  # Gap within a feature, default is 30 (i.e. 30 consecutive scans without signal), integer
    config.ppr = 0.7  # Peak-peak correlation threshold for feature grouping, default is 0.7

    # Parameters for feature alignment
    config.align_mz_tol = align_mz_tol  # m/z tolerance for MS1, default is 0.01
    config.align_rt_tol = align_rt_tol  # RT tolerance, default is 0.2
    config.alignment_drop_by_fill_pct_ratio = alignment_drop_by_fill_pct_ratio  # Drop by fill percentage ratio, default is 0.1
    config.run_rt_correction = run_rt_correction  # Whether to perform RT correction, default is True
    config.min_scan_num_for_alignment = 10  # Minimum scan number a feature to be aligned, default is 6

    # Parameters for feature annotation
    config.ms1id_library_path = ms1id_library_path
    config.library_search_mztol = library_search_mztol  # m/z tolerance for MS1 ID, default is 0.05
    config.ms2id_library_path = ms2id_library_path
    config.ms2_sim_tol = 0.7  # MS2 similarity tolerance, default is 0.7

    # Parameters for normalization
    config.run_normalization = run_normalization  # Whether to normalize the data, default is False
    config.normalization_method = "pqn"  # Normalization method, default is "pqn" (probabilistic quotient normalization), character string

    # Parameters for output
    config.output_single_file = True  # Whether to output the processed individual files to a csv file
    config.output_aligned_file = True  # Whether to output aligned features to a csv file

    # Statistical analysis
    config.run_statistics = False  # Whether to perform statistical analysis

    # Network analysis
    config.run_network = False  # Whether to perform network analysis

    # Visualization
    config.plot_bpc = False  # Whether to plot base peak chromatogram
    config.plot_ms2 = False  # Whether to plot mirror plots for MS2 matching

    return config


def feature_detection(file_name, params=None,
                      cal_g_score=False, cal_a_score=False,
                      anno_isotope=True, anno_adduct=True, anno_in_source_fragment=False,
                      ms2_library_path=None, cut_roi=True):
    """
    Untargeted feature detection from a single file (.mzML or .mzXML).

    Returns: An MSData object containing the processed data.
    """
    print("=========================================")
    print("Processing file: " + file_name)

    # create a MSData object
    d = MSData()

    # set parameters
    ms_type, ion_mode, centroid = find_ms_info(file_name)
    if not centroid:
        print("File: " + file_name + " is not centroided and skipped.")
        return None

    # if params is None, use the default parameters
    if params is None:
        params = Params()
        params.set_default(ms_type, ion_mode)

    if ms2_library_path is not None:
        params.ms2id_library_path = ms2_library_path

    print(f'Reading raw data from {file_name}...')
    # read raw data
    d.read_raw_data(file_name, params)

    print(f"Detecting features from {file_name}...")
    # detect region of interests (ROIs)
    d.find_rois()

    # cut ROIs
    if cut_roi:
        d.cut_rois()

    print(f"Summarizing features from {file_name}...")
    # label short ROIs, find the best MS2, and sort ROIs by m/z
    d.summarize_roi(cal_g_score=cal_g_score, cal_a_score=cal_a_score)

    if anno_isotope:
        print(f"Annotating isotopes from {file_name}...")
        annotate_isotope(d, mz_tol=0.015, rt_tol=params.peak_cor_rt_tol)
    # if anno_in_source_fragment:
    #     annotate_in_source_fragment(d)
    if anno_adduct:
        print(f"Annotating adducts from {file_name}...")
        annotate_adduct(d, mz_tol=0.01, rt_tol=params.peak_cor_rt_tol)

    if params.ms1_id:
        print(f"Calculating peak-peak correlations for {file_name}...")
        # calc peak-peak correlations for feature groups and output
        ppc_matrix = calc_all_ppc(d, rt_tol=params.peak_cor_rt_tol,
                                  roi_min_length=params.roi_min_length,
                                  save=False)

        print(f"Generating and saving pseudo MS2 spectra for {file_name}...")
        # generate pseudo ms2 spec, for ms1_id
        pseudo_ms2_spectra = generate_pseudo_ms2(d, ppc_matrix,
                                                 mz_tol=params.mz_tol_ms1,
                                                 min_ppc=params.min_ppc,
                                                 min_cluster_size=params.ms1id_min_matched_peak,
                                                 roi_min_length=params.roi_min_length,
                                                 save_dir=params.single_file_dir)
        del ppc_matrix

        print(f"Performing MS1 ID annotation for {file_name}...")
        # perform rev cos search
        ms1_id_annotation(pseudo_ms2_spectra, params.ms1id_library_path,
                          mz_tol=params.library_search_mztol,
                          ion_mode=params.ion_mode,
                          max_prec_rel_int_in_other_ms2=params.ms1id_max_prec_rel_int_in_other_ms2,
                          score_cutoff=params.ms1id_score_cutoff,
                          min_matched_peak=params.ms1id_min_matched_peak,
                          min_spec_usage=params.ms1id_min_spec_usage,
                          save=True,
                          save_path=os.path.join(params.single_file_dir,
                                                 os.path.basename(file_name).split(".")[
                                                     0] + "_pseudoMS2_annotated.pkl"))
        del pseudo_ms2_spectra

    # output single file to a txt file
    d.output_single_file()

    return d



import pickle

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np


def plot_single_spectrum(mz_arr, int_arr, fig_size=(3.6, 1.5),
                         peak_int_power=0.5,
                         intensity_threshold_label=0, label_decimals=2,
                         max_x=None,
                         peak_color='0.2',
                         save=False,
                         output_name='single_spectrum.svg'):
    plt.figure(figsize=fig_size)

    int_arr = np.power(int_arr, peak_int_power)

    # Normalize intensities
    max_intensity = max(int_arr[mz_arr < max_x])
    int_arr = int_arr / max_intensity * 100

    # Set axis limits
    if max_x is None:
        max_x = np.max(mz_arr) * 1.05  # Add 5% margin

    # Plot peaks
    plt.vlines(mz_arr, 0, int_arr, color=peak_color, linewidth=1.5)

    plt.xlim(0, max_x)
    plt.ylim(0, 110)

    # Add labels for peaks above threshold
    for mz, intensity in zip(mz_arr, int_arr):
        if intensity >= intensity_threshold_label and mz < max_x:
            plt.text(mz, intensity + 0.1, f'{mz:.{label_decimals}f}', ha='center', va='bottom', fontsize=8,
                     fontname='Arial')

    plt.xlabel(r'$\mathit{m/z}$', fontname='Arial', fontsize=12, labelpad=2, color='0.2')
    plt.ylabel('Intensity (%)', fontname='Arial', fontsize=12, labelpad=2, color='0.2')

    # Remove top and right spines
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    # Set the color of the frame (axes spines)
    for spine in plt.gca().spines.values():
        spine.set_color('0.5')

    # Remove grid
    plt.grid(False)

    # Use Arial font for tick labels
    plt.xticks(fontname='Arial', fontsize=10)
    plt.yticks(fontname='Arial', fontsize=10)

    # Adjust tick parameters
    plt.tick_params(axis='x', pad=4, length=2, color='0.5')
    plt.tick_params(axis='y', pad=2.5, length=2, color='0.5')

    # Set the color of tick labels to 0.2
    plt.gca().tick_params(axis='both', colors='0.2')

    # Modify y-axis ticks and labels
    yticks = plt.gca().get_yticks()
    yticks = [y for y in yticks if y <= 100]
    plt.gca().set_yticks(yticks)
    plt.gca().set_yticklabels([str(int(y)) for y in yticks])

    # Save the plot
    plt.tight_layout()
    if save:
        plt.savefig(output_name, transparent=True)
    plt.show()
    plt.close()


def create_mirror_plot(mz_arr, int_arr, mz_arr_ref, int_arr_ref, fig_size=(3.6, 2.2),
                       ms2_tol=0.05, peak_int_power=0.5,
                       intensity_threshold_label=0, label_decimals=2,
                       max_x=None,
                       matched_peak_color='red',
                       up_spec_peak_color='0.6', down_spec_peak_color='0.2',
                       show_unmatched_peaks=True,
                       save=False,
                       output_name='demo.svg'):
    plt.figure(figsize=fig_size)

    int_arr = np.power(int_arr, peak_int_power)
    int_arr_ref = np.power(int_arr_ref, peak_int_power)

    # Normalize intensities
    max_intensity = max(int_arr[mz_arr < max_x])
    int_arr = int_arr / max_intensity * 100

    max_intensity_ref = max(int_arr_ref[mz_arr_ref < max_x])
    int_arr_ref = int_arr_ref / max_intensity_ref * 100

    # Find matched peaks
    matched_indices = []
    matched_indices_ref = []
    for i, mz in enumerate(mz_arr):
        matches = np.where(np.abs(mz_arr_ref - mz) <= ms2_tol)[0]
        if matches.size > 0:
            print(f'{mz:.4f} matched to {mz_arr_ref[matches]}')
            matched_indices.extend([i] * len(matches))
            matched_indices_ref.extend(matches)

    # Set axis limits and labels
    if max_x is None:
        max_x = max(np.max(mz_arr), np.max(mz_arr_ref)) * 1.05  # Add 5% margin

    # Plot unmatched peaks
    if show_unmatched_peaks:
        plt.vlines(mz_arr, 0, int_arr, color=up_spec_peak_color, linewidth=1.5)
        plt.vlines(mz_arr_ref, 0, -int_arr_ref, color=down_spec_peak_color, linewidth=1.5)

    # Plot matched peaks
    plt.vlines(mz_arr[matched_indices], 0, int_arr[matched_indices], color=matched_peak_color, linewidth=1.5)
    plt.vlines(mz_arr_ref[matched_indices_ref], 0, -int_arr_ref[matched_indices_ref], color=matched_peak_color,
               linewidth=1.5)

    plt.xlim(0, max_x)
    plt.ylim(-110, 110)

    # Add labels for peaks above threshold
    for mz, intensity in zip(mz_arr, int_arr):
        if intensity >= intensity_threshold_label and mz < max_x:
            plt.text(mz, intensity + 0.1, f'{mz:.{label_decimals}f}', ha='center', va='bottom', fontsize=8,
                     fontname='Arial')

    for mz, intensity in zip(mz_arr_ref, int_arr_ref):
        if intensity >= intensity_threshold_label and mz < max_x:
            plt.text(mz, -intensity - 2, f'{mz:.{label_decimals}f}', ha='center', va='top', fontsize=8,
                     fontname='Arial')

    plt.xlabel(r'$\mathit{m/z}$', fontname='Arial', fontsize=12, labelpad=2, color='0.2')
    plt.ylabel('Intensity (%)', fontname='Arial', fontsize=12, labelpad=2, color='0.2')

    plt.gca().xaxis.set_visible(False)

    # Remove top and right spines
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['bottom'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    # Set the color of the frame (axes spines)
    for spine in plt.gca().spines.values():
        spine.set_color('0.5')

    # Remove grid
    plt.grid(False)

    # Add line at y=0
    plt.axhline(y=0, color='0.5', linewidth=1)

    # Use Helvetica font for tick labels
    # plt.xticks(fontname='Arial', fontsize=10)
    plt.yticks(fontname='Arial', fontsize=10)

    # Adjust tick parameters
    # plt.tick_params(axis='x', pad=4, length=2, color='0.5')
    plt.tick_params(axis='y', pad=2.5, length=2, color='0.5')

    # Set the color of tick labels to 0.2
    plt.gca().tick_params(axis='both', colors='0.2')

    # Modify y-axis ticks and labels
    yticks = plt.gca().get_yticks()
    yticks = [y for y in yticks if abs(y) <= 100]
    plt.gca().set_yticks(yticks)
    plt.gca().set_yticklabels([str(abs(int(y))) for y in yticks])

    # Save the plot
    plt.tight_layout()
    if save:
        plt.savefig(output_name, transparent=True)
    plt.show()
    plt.close()


def get_ref_spec(std_k0, std_k10, annotation, precursor_type, db_name):
    if db_name == 'std_k0.pkl':
        eng = std_k0
    elif db_name == 'std_k10.pkl':
        eng = std_k10
    else:
        raise ValueError(f'Unknown database name: {db_name}')

    for spec in eng:
        if spec['name'] == annotation and spec['precursor_type'] == precursor_type:
            mz_arr_ref = spec['peaks'][:, 0]
            int_arr_ref = spec['peaks'][:, 1]
            return mz_arr_ref, int_arr_ref

    raise ValueError(f'No reference spectrum found for {annotation} ({precursor_type})')


def demo(name, index, file='mixture_Bile_acids_20eV'):
    df = pd.read_csv(
        f'output_all/{file}_feature_table.tsv', sep='\t', low_memory=False)

    df = df[df['MS1_similarity'].notnull()]
    df = df[df['MS1_matched_peak'] >= 4].reset_index(drop=True)
    df = df.sort_values('MS1_similarity', ascending=False).reset_index(drop=True)

    if name is not None:
        idx = df[df['MS1_annotation'] == name].index[index]
    else:
        idx = index

    print(df.iloc[idx]['MS1_annotation'])
    print(df.iloc[idx]['MS1_precursor_type'])
    print(df.iloc[idx]['m/z'])
    print(df.iloc[idx]['MS1_similarity'])
    print(df.iloc[idx]['MS1_matched_peak'])

    ms1 = df.iloc[idx]['pseudo_ms2']
    mz_arr = []
    int_arr = []
    for p in ms1.split(';'):
        p = p.strip()
        if not p:
            continue
        mz, inten = p.split(' ')
        mz_arr.append(float(mz))
        int_arr.append(float(inten))

    mz_arr = np.array(mz_arr)
    int_arr = np.array(int_arr)

    # load search engine
    with open('/data/std_mix/std_k0.pkl', 'rb') as file:
        std_k0 = pickle.load(file)
    with open('/data/std_mix/std_k10.pkl', 'rb') as file:
        std_k10 = pickle.load(file)

    # mz_arr_ref, int_arr_ref = get_ref_spec(std_k0, std_k10,
    #                                        df.iloc[idx]['MS1_annotation'],
    #                                        df.iloc[idx]['MS1_precursor_type'],
    #                                        df.iloc[idx]['MS1_db_name'])

    mz_arr_ref, int_arr_ref = get_ref_spec(std_k0, std_k10,
                                           df.iloc[idx]['MS1_annotation'],
                                           df.iloc[idx]['MS1_precursor_type'],
                                           'std_k0.pkl')

    # Create a mask for mz values in mz_arr that have a close match in mz_arr_ref
    mask = np.any(np.abs(mz_arr[:, np.newaxis] - mz_arr_ref) <= 0.02, axis=1)

    print('len(mz_arr):', len(mz_arr))

    # Apply the mask to both mz_arr and int_arr
    mz_arr = mz_arr[mask]
    print('len(mz_arr):', len(mz_arr))
    int_arr = int_arr[mask]

    alpha = np.array(mz_arr_ref) / float(df.iloc[idx]['m/z']) * 10
    int_arr_ref = np.array(int_arr_ref) * np.exp(alpha)

    create_mirror_plot(np.array(mz_arr), np.array(int_arr),
                       np.array(mz_arr_ref), np.array(int_arr_ref),
                       #fig_size=(12, 6),
                       ms2_tol=0.02, peak_int_power=1,
                       intensity_threshold_label=101,
                       label_decimals=2,
                       # max_x=float(df.iloc[idx]['m/z']) * 0.99,
                       max_x=float(df.iloc[idx]['m/z']) * 1.02,
                       matched_peak_color='#F08976',
                       # matched_peak_color='0.65',
                       up_spec_peak_color='0.75', down_spec_peak_color='0.65',
                       show_unmatched_peaks=True,
                       save=True,
                       output_name='demo2.svg')

    # plot_single_spectrum(mz_arr, int_arr,
    #                      fig_size=(3.6, 1.7),
    #                      peak_int_power=0.5,
    #                      intensity_threshold_label=101, label_decimals=2,
    #                      max_x=float(df.iloc[idx]['m/z']) * 1.05,
    #                      peak_color='#aa3e53',
    #                      save=True,
    #                      output_name='sulfadimethoxine3.svg')


if __name__ == '__main__':

    # mixture_Bile_acids_20eV
    # mixture_Drugs_20eV
    # sixmix_2_20eV

    demo(None, 0, 'mixture_Drugs_20eV')
    # demo(None, 0, 'sixmix_2_10eV')

    # demo('venlafaxine', 0, 'mixture_Drugs_20eV')
    # demo('sulfadimethoxine', 0, 'sixmix_2_20eV')
    # demo('Taurolithocholic acid', 2, 'mixture_Bile_acids_20eV')

    # cosine between two vectors
    def scale(peaks, prec_mz):
        scaling_factor = peaks[:, 0] / prec_mz * 10
        # print('scaling_factor:', scaling_factor)
        return peaks[:, 1] * np.exp(scaling_factor)


    def cosine_similarity(v1, v2):
        return np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))


    print(cosine_similarity(np.array([3, 7, 5, 4, 8, 17, 0, 0, 0, 0, 0]),
                            np.array([3, 7, 5, 4, 8, 17, 11, 4, 14, 6, 5])))



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats


def get_peak_height_ls():
    df = pd.read_pickle('data/aligned_feature_table_all.pkl')
    df['max_intensity'] = df.iloc[:, 22:28].max(axis=1)

    ms2_dda_ph_ls = df[df['similarity'].notnull()]['max_intensity'].tolist()
    ms1_dda_ph_ls = df[df['ms1_dda_inchikeys'].notnull()]['max_intensity'].tolist()
    ms1_0ev_ph_ls = df[df['ms1_0ev_inchikeys'].notnull()]['max_intensity'].tolist()
    ms1_10ev_ph_ls = df[df['ms1_10ev_inchikeys'].notnull()]['max_intensity'].tolist()
    ms1_20ev_ph_ls = df[df['ms1_20ev_inchikeys'].notnull()]['max_intensity'].tolist()

    return ms2_dda_ph_ls, ms1_dda_ph_ls, ms1_0ev_ph_ls, ms1_10ev_ph_ls, ms1_20ev_ph_ls


# Set up the plot
plt.rcParams['font.family'] = 'Arial'
fig, ax = plt.subplots(figsize=(4.85, 1.95))

# Get peak height data
peak_heights = get_peak_height_ls()

# Labels for y-axis
labels = ['MS1 (20 eV)', 'MS1 (10 eV)', 'MS1 (0 eV)', 'MS1 (DDA)', 'MS/MS (DDA)']

# Plot scatter points behind the box
for i, heights in enumerate(reversed(peak_heights)):
    y = np.random.normal(i + 1, 0.05, len(heights))
    ax.scatter(heights, y, alpha=0.15, s=2.5, color='blue', edgecolors='none', zorder=1)

# Create box plot
box_color = 'lightgrey'
box_width = 0.35
box_plot = ax.boxplot(list(reversed(peak_heights)), vert=False, patch_artist=True, labels=labels, showfliers=False,
                      widths=box_width)

# Customize box appearance
for element in ['boxes', 'fliers', 'means', 'medians']:
    plt.setp(box_plot[element], color='none')

plt.setp(box_plot['whiskers'], color='0.5')
plt.setp(box_plot['caps'], color='0.5')

for median in box_plot['medians']:
    median.set(color='white', linewidth=1.5)

for patch in box_plot['boxes']:
    patch.set_facecolor(box_color)
    patch.set_edgecolor('none')
    patch.set_alpha(0.85)

# Customize the plot
ax.set_xlabel('Peak height of annotated features', fontsize=12, fontname='Arial', labelpad=3.5, color='0.2')
ax.set_xscale('log')

# Set x-axis limits and show minor ticks
ax.set_xlim(2.45e5, 1.5e8)
ax.xaxis.set_minor_locator(plt.LogLocator(subs=np.arange(2, 10)))

# Customize ticks and remove top and right spines
ax.tick_params(axis='y', which='major', labelsize=10, pad=5, length=1.5, colors='0.2')
ax.tick_params(axis='x', which='major', labelsize=8, pad=2, length=2, colors='0.2')
ax.tick_params(axis='x', which='minor', length=1.3, colors='0.2')
ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)
ax.spines['bottom'].set_color('0.5')
ax.spines['left'].set_color('0.5')

# Set x-axis and y-axis label font to Arial and color to 0.2
for label in ax.get_xticklabels() + ax.get_yticklabels():
    label.set_fontname('Arial')
    label.set_color('0.2')

# Adjust the space between the first box and x-axis
plt.ylim(0.3, len(labels) + 0.5)

plt.tight_layout()

# Save as svg
plt.savefig('data/peak_height_vertical_boxplot.svg', transparent=True, bbox_inches='tight')

plt.show()


# Perform Mann-Whitney U test
ms2_dda_ph_ls, ms1_dda_ph_ls, ms1_0ev_ph_ls, ms1_10ev_ph_ls, ms1_20ev_ph_ls = peak_heights

comparisons = [
    ('MS/MS (DDA)', 'MS1 (DDA)', ms2_dda_ph_ls, ms1_dda_ph_ls),
    ('MS/MS (DDA)', 'MS1 (0 eV)', ms2_dda_ph_ls, ms1_0ev_ph_ls),
    ('MS/MS (DDA)', 'MS1 (10 eV)', ms2_dda_ph_ls, ms1_10ev_ph_ls),
    ('MS/MS (DDA)', 'MS1 (20 eV)', ms2_dda_ph_ls, ms1_20ev_ph_ls)
]

print("Mann-Whitney U test results:")
for group1, group2, data1, data2 in comparisons:
    statistic, p_value = stats.mannwhitneyu(data1, data2, alternative='two-sided')
    print(f"{group1} vs {group2}: p-value = {p_value:.4e}")


import pandas as pd
import numpy as np
from scipy import stats
from statsmodels.stats.multitest import multipletests
from tqdm import tqdm


def fill_zeros(row):
    row = pd.to_numeric(row, errors='coerce')
    non_zero_min = row[row > 0].min() if (row > 0).any() else 0
    fill_value = min(30000, 0.1 * non_zero_min)
    return row.fillna(fill_value).replace(0, fill_value)


def calculate_fold_change(group1_data, group2_data):
    return group1_data.mean() / group2_data.mean()


def remove_outliers_iqr(data):
    Q1 = data.quantile(0.25)
    Q3 = data.quantile(0.75)
    IQR = Q3 - Q1
    lower_bound = Q1 - 1.5 * IQR
    upper_bound = Q3 + 1.5 * IQR
    return data[(data >= lower_bound) & (data <= upper_bound)]


def perform_statistical_analysis(mode):
    df = pd.read_csv(
        f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/{mode}_aligned_feature_table_with_metadata.tsv',
        sep='\t', low_memory=False)

    diagnosis = df.iloc[0]
    sex = df.iloc[1]

    data = df.iloc[2:]

    metadata_cols = list(data.columns[:22]) + list(data.columns[-8:])
    data_cols = [col for col in data.columns if col not in metadata_cols]

    data[data_cols] = data[data_cols].apply(fill_zeros, axis=1)

    comparisons = [
        ('nonIBD', 'CD', None),
        ('nonIBD', 'UC', None),
        ('nonIBD', 'CD', 'male'),
        ('nonIBD', 'CD', 'female'),
        ('nonIBD', 'UC', 'male'),
        ('nonIBD', 'UC', 'female')
    ]

    results = {}

    for comparison in comparisons:
        group1, group2, gender = comparison

        if gender:
            cols1 = [col for col in data_cols if diagnosis[col] == group1 and sex[col].lower() == gender.lower()]
            cols2 = [col for col in data_cols if diagnosis[col] == group2 and sex[col].lower() == gender.lower()]
        else:
            cols1 = [col for col in data_cols if diagnosis[col] == group1]
            cols2 = [col for col in data_cols if diagnosis[col] == group2]

        print(f"Comparison: {comparison}")
        print(f"Number of columns in group1: {len(cols1)}")
        print(f"Number of columns in group2: {len(cols2)}")

        p_values = []
        fold_changes = []
        for _, row in tqdm(data[data_cols].iterrows(), desc=f"Processing {comparison}"):
            group1_data = row[cols1].astype(float)
            group2_data = row[cols2].astype(float)

            # Remove outliers for each group separately
            group1_data_clean = remove_outliers_iqr(group1_data)
            group2_data_clean = remove_outliers_iqr(group2_data)

            if len(group1_data_clean) > 0 and len(group2_data_clean) > 0:
                stat, p = stats.mannwhitneyu(group1_data_clean, group2_data_clean, alternative='two-sided')
                fc = calculate_fold_change(group1_data_clean, group2_data_clean)
                p_values.append(p)
                fold_changes.append(fc)
            else:
                p_values.append(np.nan)
                fold_changes.append(np.nan)

        valid_p_values = [p for p in p_values if not np.isnan(p)]
        if valid_p_values:
            _, p_corrected, _, _ = multipletests(valid_p_values, method='fdr_bh')
            p_corrected_full = [p_corrected[valid_p_values.index(p)] if not np.isnan(p) else np.nan for p in p_values]
        else:
            p_corrected_full = [np.nan] * len(p_values)

        comparison_name = f"{group1}_vs_{group2}{'_' + gender if gender else ''}"
        results[f"{comparison_name}_p_value"] = p_values
        results[f"{comparison_name}_p_value_corrected"] = p_corrected_full
        results[f"{comparison_name}_fold_change"] = fold_changes

    result_df = pd.DataFrame(results, index=data.index)
    combined_df = pd.concat([data, result_df], axis=1)
    cols_order = metadata_cols + data_cols + list(results.keys())
    combined_df = combined_df[cols_order]

    return combined_df


def stats_main():
    modes = ['hilic_pos', 'hilic_neg', 'c8_pos', 'c18_neg']

    for mode in modes:
        print(f"\nProcessing {mode}...")
        results = perform_statistical_analysis(mode)
        results.to_csv(
            f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/{mode}_statistical_analysis_with_outlier_removal.tsv',
            sep='\t', index=False)


def merge_all_modes():
    """
    merge all modes into one DataFrame, only shared columns are kept
    """
    modes = ['hilic_pos', 'hilic_neg', 'c8_pos', 'c18_neg']

    dfs = []
    for mode in modes:
        df = pd.read_csv(
            f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/{mode}_statistical_analysis_with_outlier_removal.tsv',
            sep='\t', low_memory=False)
        print(f"Number of columns in {mode}: {len(df.columns)}")
        df['mode'] = mode
        dfs.append(df)

    common_columns = set(dfs[0].columns[30:-19])
    for df in dfs[1:]:
        common_columns = common_columns.intersection(set(df.columns[30:-19]))

    common_columns = list(common_columns)
    common_columns = dfs[0].columns[:30].tolist() + dfs[0].columns[-19:].tolist() + common_columns

    print(f"Number of common columns: {len(common_columns)}")

    dfs_common = [df[common_columns] for df in dfs]

    combined_df = pd.concat(dfs_common, axis=0)

    print(f"Shape of combined DataFrame: {combined_df.shape}")

    combined_df.to_csv('all_modes_statistical_analysis_with_outlier_removal.tsv',
                       sep='\t', index=False)


if __name__ == '__main__':
    stats_main()
    merge_all_modes()


import pickle
import re

import numpy as np

from ms1_id.lcms.centroid_data import centroid_spectrum_for_search


def feature_annotation(features, config, num=5):
    """
    A function to annotate features based on their MS/MS spectra and a MS/MS database.

    Parameters
    ----------
    features : list
        A list of features.
    config : Params object
        The parameters for the workflow.
    num : int
        The number of top MS/MS spectra to search.
    """

    # load the MS/MS database
    search_eng = pickle.load(open(config.ms2id_library_path, 'rb'))

    library_search_mztol = min(config.library_search_mztol, 0.05)  # indexed library mz_tol is 0.05

    for f in features:
        if len(f.ms2_seq) == 0:
            continue
        parsed_ms2 = []
        for ms2 in f.ms2_seq:
            peaks = _extract_peaks_from_string(ms2)
            parsed_ms2.append(peaks)
        # sort parsed ms2 by summed intensity
        parsed_ms2 = sorted(parsed_ms2, key=lambda x: np.sum(x[:, 1]), reverse=True)
        parsed_ms2 = parsed_ms2[:num]

        all_valid_matches = []

        for peaks in parsed_ms2:
            peaks = centroid_spectrum_for_search(peaks, width_da=0.05 * 2.015)

            if len(peaks) < config.ms2id_min_matched_peak:
                continue

            search_result = search_eng.search(
                precursor_mz=f.mz,
                peaks=peaks,
                ms1_tolerance_in_da=library_search_mztol,
                ms2_tolerance_in_da=library_search_mztol,
                method="identity",
                precursor_ions_removal_da=0.5,
                noise_threshold=0.0,
                min_ms2_difference_in_da=library_search_mztol * 2.02,
                reverse=False
            )
            score_arr, matched_peak_arr, spec_usage_arr = search_result['identity_search']

            # Find indices that pass all filters
            valid_indices = np.where((matched_peak_arr >= config.ms2id_min_matched_peak) &
                                     (score_arr >= config.ms2id_score_cutoff))[0]

            for idx in valid_indices:
                potential_match = search_eng[idx]
                potential_match = {k.lower(): v for k, v in potential_match.items()}

                if potential_match.get('ion_mode', '') == config.ion_mode:
                    all_valid_matches.append({
                        'match': potential_match,
                        'score': score_arr[idx],
                        'matched_peak': matched_peak_arr[idx],
                        'ms2': peaks
                    })

        # Select the match with the highest score
        if all_valid_matches:
            best_match = max(all_valid_matches, key=lambda x: x['score'])

            f.annotation = best_match['match'].get('name', None)
            f.search_mode = 'identity_search'
            f.similarity = best_match['score']
            f.matched_peak_number = best_match['matched_peak']
            f.smiles = best_match['match'].get('smiles', None)
            f.inchikey = best_match['match'].get('inchikey', None)
            f.matched_ms2 = _convert_peaks_to_string(best_match['match']['peaks'])
            f.formula = best_match['match'].get('formula', None)
            f.adduct_type = best_match['match'].get('precursor_type')
            f.best_ms2 = _convert_peaks_to_string(best_match['ms2'])
            f.collision_energy = best_match['match'].get('collision_energy', None)
            f.precursor_type = best_match['match'].get('precursor_type', None)
        else:
            f.best_ms2 = _convert_peaks_to_string(parsed_ms2[0])

    return features


def annotate_rois(d, ms2id_score_cutoff=0.8, ms2id_min_matched_peak=6, ion_mode='positive'):
    """
    A function to annotate rois based on their MS/MS spectra and a MS/MS database.

    Parameters
    ----------
    d : MSData object
        MS data.
    """

    # load the MS/MS database
    search_eng = pickle.load(open(d.params.ms2id_library_path, 'rb'))

    ms2_tol = min(d.params.mz_tol_ms2, 0.05)  # indexed library mz_tol is 0.05

    for f in d.rois:
        f.annotation = None
        f.similarity = None
        f.matched_peak_number = None
        f.smiles = None
        f.inchikey = None
        f.matched_precursor_mz = None
        f.matched_peaks = None
        f.formula = None
        f.precursor_type = None
        f.collision_energy = None

        if f.best_ms2 is not None:

            peaks = f.best_ms2.peaks

            # centroid
            peaks = centroid_spectrum_for_search(peaks, width_da=0.05 * 2.015)

            if len(peaks) < ms2id_min_matched_peak:
                continue

            search_result = search_eng.search(
                precursor_mz=f.mz,
                peaks=peaks,
                ms1_tolerance_in_da=d.params.mz_tol_ms1,
                ms2_tolerance_in_da=ms2_tol,
                method="identity",
                precursor_ions_removal_da=0.5,
                noise_threshold=0.0,
                min_ms2_difference_in_da=ms2_tol * 2.2,
                reverse=False
            )
            score_arr, matched_peak_arr, spec_usage_arr = search_result['identity_search']

            # Find indices that pass all filters
            valid_indices = np.where((matched_peak_arr >= ms2id_min_matched_peak) &
                                     (score_arr >= ms2id_score_cutoff))[0]

            all_valid_matches = []

            for idx in valid_indices:
                potential_match = search_eng[idx]
                potential_match = {k.lower(): v for k, v in potential_match.items()}

                if potential_match.get('ion_mode', '') == ion_mode:
                    all_valid_matches.append({
                        'match': potential_match,
                        'score': score_arr[idx],
                        'matched_peak': matched_peak_arr[idx]
                    })

            # Select the match with the highest score
            if all_valid_matches:
                best_match = max(all_valid_matches, key=lambda x: x['score'])

                f.annotation = best_match['match'].get('name', None)
                f.similarity = best_match['score']
                f.matched_peak_number = best_match['matched_peak']
                f.smiles = best_match['match'].get('smiles', None)
                f.inchikey = best_match['match'].get('inchikey', None)
                f.matched_precursor_mz = best_match['match'].get('precursor_mz', None)
                f.matched_peaks = best_match['match'].get('peaks', None)
                f.formula = best_match['match'].get('formula', None)
                f.precursor_type = best_match['match'].get('precursor_type', None)
                f.collision_energy = best_match['match'].get('collision_energy', None)


def _extract_peaks_from_string(ms2):
    """
    Extract peaks from MS2 spectrum.

    Parameters
    ----------
    ms2 : str
        MS2 spectrum in string format.

    Example
    ----------

    """

    # Use findall function to extract all numbers matching the pattern
    numbers = re.findall(r'\d+\.\d+', ms2)

    # Convert the extracted numbers from strings to floats
    numbers = [float(num) for num in numbers]

    numbers = np.array(numbers).reshape(-1, 2)

    return numbers


def _convert_peaks_to_string(peaks):
    """
    Convert peaks to string format.

    Parameters
    ----------
    peaks : numpy.array
        Peaks in numpy array format.

    Example
    ----------

    """

    ms2 = ""
    for i in range(len(peaks)):
        ms2 += str(np.round(peaks[i, 0], decimals=4)) + ";" + str(np.round(peaks[i, 1], decimals=4)) + "|"
    ms2 = ms2[:-1]

    return ms2



import pandas as pd


df = pd.read_csv('/Users/shipei/Documents/projects/ms1_id/imaging/mouse_body/wb xenograft in situ metabolomics test - rms_corrected/ms1_id_annotations_all.tsv', sep='\t', low_memory=False)

df = df[~pd.isnull(df['inchikey'])].reset_index(drop=True)

print(f"Number of unique inchikeys: {df['inchikey'].nunique()}")



"""
compare mol. network w/ or w/o pseudo MS2
"""

import pandas as pd
import numpy as np
from tqdm import tqdm
import networkx as nx


def dda_mn():
    # read feature table
    df = pd.read_csv(f'/data/nist/dda_ms2/aligned_feature_table.tsv',
                     sep='\t', low_memory=False)

    with open(f'data/nist_DDA_MS2.mgf', 'w') as f:
        for _, row in tqdm(df.iterrows(), total=len(df), desc="Writing MGF"):
            if pd.notnull(row['MS2']):

                pms2 = row['MS2'].split('|')
                pms2 = [x.split(';') for x in pms2]
                pms2 = [[float(x) for x in y] for y in pms2]
                pms2 = np.array(pms2, dtype=np.float32)

                # remove ions larger than precursor
                pms2 = pms2[pms2[:, 0] < row['m/z'] + 0.1]

                mzs = pms2[:, 0]
                intensities = pms2[:, 1]

                f.write('BEGIN IONS\n')
                f.write(f'NAME={row["annotation"]}\n')
                f.write(f'PEPMASS={row["m/z"]}\n')
                f.write(f'RTINSECONDS={row["RT"] * 60}\n')
                f.write(f'SCANS={round(row["ID"])}\n')

                for mz, intensity in zip(mzs, intensities):
                    f.write(f'{mz} {intensity}\n')
                f.write('END IONS\n\n')


def pms2_mn(mode='fullscan_0ev'):
    # read feature table
    df = pd.read_csv(f'/Users/shipei/Documents/projects/ms1_id/data/nist/{mode}/aligned_feature_table.tsv',
                     sep='\t', low_memory=False)

    with open(f'data/nist_{mode}_pMS2.mgf', 'w') as f:
        for _, row in tqdm(df.iterrows(), total=len(df), desc="Writing MGF"):
            if pd.notnull(row['pseudo_ms2']):

                pms2 = row['pseudo_ms2'].split(';')
                pms2 = [x.split(' ') for x in pms2][:-1]
                pms2 = [[float(x) for x in y] for y in pms2]
                pms2 = np.array(pms2, dtype=np.float32)

                # remove ions larger than precursor
                pms2 = pms2[pms2[:, 0] < row['m/z'] + 0.1]

                mzs = pms2[:, 0]
                intensities = pms2[:, 1]

                f.write('BEGIN IONS\n')
                f.write(f'NAME={row["MS1_annotation"]}\n')
                f.write(f'PEPMASS={row["m/z"]}\n')
                f.write(f'RTINSECONDS={row["RT"] * 60}\n')
                f.write(f'SCANS={round(row["ID"])}\n')

                for mz, intensity in zip(mzs, intensities):
                    f.write(f'{mz} {intensity}\n')
                f.write('END IONS\n\n')


def add_annotations_to_network_file():
    G = nx.read_graphml('data/mn/fullscan0ev_network.graphml')

    # read feature table
    dda_df = pd.read_csv(f'/data/nist/dda_ms2/aligned_feature_table.tsv',
                         sep='\t', low_memory=False)
    dda_df.loc[dda_df['similarity'] < 0.8, 'annotation'] = None

    ms1_df = pd.read_csv(f'/data/nist/fullscan_0ev/aligned_feature_table.tsv',
                         sep='\t', low_memory=False)
    ms1_df.loc[ms1_df['MS1_similarity'] < 0.8, 'MS1_annotation'] = None

    # Iterate through all nodes in the graph
    for node, data in G.nodes(data=True):
        mz = data['mz']
        rt = data['rt']

        # find in dda_df
        dda_row = dda_df[(dda_df['m/z'].between(mz-0.01, mz+0.01)) & (dda_df['RT'].between(rt-0.05, rt+0.05))]

        if len(dda_row) > 0:
            # sort by matched_score
            dda_row = dda_row.sort_values('similarity', ascending=False).iloc[0]
            data['DDA_feature'] = True

            # if annotated
            if not pd.isnull(dda_row['annotation']):
                data['MS2_annotation'] = dda_row['annotation']
                data['DDA_annotated'] = True
            else:
                data['MS2_annotation'] = ''
                data['DDA_annotated'] = False
        else:
            data['DDA_feature'] = False
            data['MS2_annotation'] = ''
            data['DDA_annotated'] = False

        # find in ms1_df
        ms1_row = ms1_df[(ms1_df['m/z'].between(mz-0.01, mz+0.01)) & (ms1_df['RT'].between(rt-0.05, rt+0.05))]
        if len(ms1_row) > 0:
            # sort by RT difference
            ms1_row['RT_diff'] = np.abs(ms1_row['RT'] - rt)
            ms1_row = ms1_row.sort_values('RT_diff', ascending=True).iloc[0]

            data['MS1_feature'] = True

            # if annotated
            if not pd.isnull(ms1_row['MS1_annotation']):
                data['MS1_annotation'] = ms1_row['MS1_annotation']
                data['MS1_annotated'] = True
            else:
                data['MS1_annotation'] = ''
                data['MS1_annotated'] = False
        else:
            data['MS1_feature'] = False
            data['MS1_annotation'] = ''
            data['MS1_annotated'] = False

    # Save the modified graph to a new GraphML file
    nx.write_graphml(G, 'data/mn/fullscan0ev_network_annotations.graphml')


if __name__ == '__main__':
    #########
    # generate mgf files for mol. network
    # dda_mn()
    # pms2_mn('fullscan_0ev')

    #########
    add_annotations_to_network_file()



import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
import math
import pandas as pd
from requests import get
from json import loads
import time
from functools import wraps


def retry_on_exception(max_attempts=5, delay=0.1):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempts = 0
            while attempts < max_attempts:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    attempts += 1
                    if attempts == max_attempts:
                        raise e
                    time.sleep(delay)
            return None
        return wrapper
    return decorator


def usi_to_spec(usi):
    # get spectrum from USI
    url = 'https://metabolomics-usi.gnps2.org/json/?usi1=' + usi
    response = get(url, timeout=30)
    json_data = loads(response.text)
    ms2_mz = np.array(json_data['peaks'])[:, 0]
    ms2_int = np.array(json_data['peaks'])[:, 1]
    return ms2_mz, ms2_int, json_data['precursor_mz']


@retry_on_exception(max_attempts=5, delay=0.1)
def gnps_id_to_spec(gnps_id):
    usi = f'mzspec:GNPS:GNPS-LIBRARY:accession:{gnps_id}'
    return usi_to_spec(usi)

def create_single_plot(ax,
                       mz_arr, int_arr, prec_mz,
                       mz_arr_ref, int_arr_ref,
                       max_x, mz_range=(100, 1200),
                       ref_scale=0,
                       query_matched_only=False,
                       ms2_tol=0.05, peak_int_power=1.0,
                       plot_info=None, plot_title=None):
    int_arr = np.power(int_arr, peak_int_power)
    int_arr_ref = np.power(int_arr_ref, peak_int_power)

    # mz range
    mask = (mz_arr >= mz_range[0]) & (mz_arr <= min(mz_range[1], max_x))
    mz_arr = mz_arr[mask]
    int_arr = int_arr[mask]
    mask = (mz_arr_ref >= mz_range[0]) & (mz_arr_ref <= min(mz_range[1], max_x))
    mz_arr_ref = mz_arr_ref[mask]
    int_arr_ref = int_arr_ref[mask]

    # Scale ref MS2
    factor = mz_arr_ref / prec_mz
    int_arr_ref = int_arr_ref * np.exp(ref_scale * factor)

    # Find matched peaks
    matched_indices = []
    matched_indices_ref = []
    for i, mz in enumerate(mz_arr):
        matches = np.where(np.abs(mz_arr_ref - mz) <= ms2_tol)[0]
        if matches.size > 0:
            matched_indices.extend([i] * len(matches))
            matched_indices_ref.extend(matches)

    # Normalize intensities for pseudo MS2
    if matched_indices:
        max_matched_intensity = max(int_arr[matched_indices])
        int_arr = int_arr / max_matched_intensity * 100
    else:
        # If no matches, normalize by the overall maximum
        int_arr = int_arr / max(int_arr) * 100

    # Normalize intensities for ref MS2
    max_intensity_ref = max(int_arr_ref)
    int_arr_ref = int_arr_ref / max_intensity_ref * 100

    # Plot unmatched peaks
    if not query_matched_only:
        ax.vlines(mz_arr, 0, int_arr, color='0.7', linewidth=0.6)
    ax.vlines(mz_arr_ref, 0, -int_arr_ref, color='0.7', linewidth=0.6)

    # Plot matched peaks
    ax.vlines(mz_arr[matched_indices], 0, int_arr[matched_indices], color='#F08976', linewidth=0.7)
    ax.vlines(mz_arr_ref[matched_indices_ref], 0, -int_arr_ref[matched_indices_ref], color='#F08976', linewidth=0.7)

    ax.set_xlim(0, max_x)
    ax.set_ylim(-120, 120)

    ax.set_xlabel(r'$\mathit{m/z}$', fontname='Arial', fontsize=8, labelpad=2, color='0.2')
    ax.set_ylabel('Intensity (%)', fontname='Arial', fontsize=8, labelpad=2, color='0.2')

    # Add "Pseudo MS2" and "Ref. MS2" labels
    ax.text(0.02, 0.95, 'Pseudo MS2', transform=ax.transAxes, ha='left', va='top', fontsize=7, color='0.2')
    ax.text(0.02, 0.20, 'Ref. MS2', transform=ax.transAxes, ha='left', va='bottom', fontsize=7, color='0.2')

    # Add plot information above "Ref. MS2" label
    if plot_info:
        ax.text(0.02, 0.30, plot_info, transform=ax.transAxes, ha='left', va='bottom', fontsize=5.5, color='0.2')

    ax.xaxis.set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['right'].set_visible(False)
    for spine in ax.spines.values():
        spine.set_color('0.5')

    ax.grid(False)
    ax.axhline(y=0, color='0.5', linewidth=0.5)

    ax.tick_params(axis='y', pad=2, length=2, color='0.5')
    ax.tick_params(axis='both', colors='0.2')

    yticks = ax.get_yticks()
    yticks = [y for y in yticks if abs(y) <= 100]
    ax.set_yticks(yticks)
    ax.set_yticklabels([str(abs(int(y))) for y in yticks], fontsize=6)

    # Add plot title if provided
    if plot_title:
        ax.set_title(plot_title, fontsize=8, pad=2, wrap=True)


def plot_multiple_figures(annotation_file, output_pdf_name='multiple_plots.pdf',
                          min_score=0.7, min_matched_peak=3, min_spectral_usage=0.15, peak_int_power=1.0,
                          rows=5, cols=2):
    df = pd.read_csv(annotation_file, sep='\t', low_memory=False)

    df = df.sort_values(['mz', 'matched_score'], ascending=[True, False])
    # df = df.sort_values(['spectral_usage', 'matched_score', 'matched_peak'], ascending=[False, False, False])

    df['gnps_id'] = df['db_id'].apply(lambda x: x.split('=')[1].split(';')[0])

    # Filter by score, matched peak, and spectral usage
    df = df[(df['matched_score'] >= min_score) &
            (df['matched_peak'] >= min_matched_peak) &
            (df['spectral_usage'] >= min_spectral_usage)].reset_index(drop=True)

    print(f"Loaded {len(df)} annotations from {annotation_file}")

    with PdfPages(output_pdf_name) as pdf:
        num_plots = len(df)
        plots_per_page = rows * (cols // 2)  # Each spectrum takes 2 columns

        for page in range(math.ceil(num_plots / plots_per_page)):
            start_idx = page * plots_per_page
            end_idx = min((page + 1) * plots_per_page, num_plots)
            plots_on_page = end_idx - start_idx

            # Calculate the number of rows needed for this page
            rows_on_page = math.ceil(plots_on_page / (cols // 2))

            # Create figure with extra space at the top of each row for titles
            fig, axs = plt.subplots(rows_on_page, cols, figsize=(cols * 6, rows_on_page * 2.3))
            fig.subplots_adjust(hspace=0.5, wspace=0.3, top=0.95)

            # Ensure axs is always a 2D array
            if rows_on_page == 1:
                axs = axs.reshape(1, -1)

            for i in range(plots_on_page):
                idx = start_idx + i
                row = i // (cols // 2)
                col = (i % (cols // 2)) * 2

                print('Processing', idx, df.iloc[idx]['name'])
                ms1 = df.iloc[idx]['pseudo_ms2']
                mz_arr = []
                int_arr = []
                for p in ms1.split(';'):
                    p = p.strip()
                    if not p:
                        continue
                    mz, inten = p.split(' ')
                    if float(inten) > 0:
                        mz_arr.append(float(mz))
                        int_arr.append(float(inten))

                mz_arr_ref, int_arr_ref, _ = gnps_id_to_spec(df.iloc[idx]['gnps_id'])
                # add precursor m/z for visualization
                if np.min(np.abs(mz_arr_ref - df.iloc[idx]['precursor_mz'])) > 0.05:
                    mz_arr_ref = np.append(mz_arr_ref, df.iloc[idx]['precursor_mz'])
                    int_arr_ref = np.append(int_arr_ref, 0.01)

                prec_mz = float(df.iloc[idx]['precursor_mz'])

                # Prepare plot information
                plot_info = f"Score: {df.iloc[idx]['matched_score']:.2f}\n" \
                            f"Matched: {df.iloc[idx]['matched_peak']} (w/o precursor)\n" \
                            f"Usage: {df.iloc[idx]['spectral_usage']:.2f}"

                # Prepare plot title
                metabolite_name = df.iloc[idx]['name']
                mz_value = df.iloc[idx]['precursor_mz']
                precursor_type = df.iloc[idx]['precursor_type']
                plot_title = f"{metabolite_name}\nm/z: {mz_value:.4f}, {precursor_type}"

                if metabolite_name == 'PC(0:0/18:0); [M+H]+ C26H55N1O7P1':
                    print('PC(0:0/18:0); [M+H]+ C26H55N1O7P1')

                # Original plot
                create_single_plot(axs[row, col], np.array(mz_arr), np.array(int_arr), prec_mz,
                                   np.array(mz_arr_ref), np.array(int_arr_ref), prec_mz + 2,
                                   ref_scale=0,
                                   peak_int_power=peak_int_power,
                                   query_matched_only=True,
                                   plot_info=plot_info, plot_title=plot_title)
                axs[row, col].set_title(plot_title, fontsize=6.5, pad=2, wrap=True)
                axs[row, col].text(0.5, 0.88, "Non-scaled", transform=axs[row, col].transAxes, ha='center', va='bottom',
                                   fontsize=6)

                # Square root plot
                create_single_plot(axs[row, col + 1], np.array(mz_arr), np.array(int_arr), prec_mz,
                                   np.array(mz_arr_ref), np.array(int_arr_ref), prec_mz + 2,
                                   ref_scale=10,
                                   peak_int_power=peak_int_power,
                                   query_matched_only=True,
                                   plot_info=plot_info, plot_title=plot_title)
                axs[row, col + 1].set_title(plot_title, fontsize=6.5, pad=2, wrap=True)
                axs[row, col + 1].text(0.5, 0.88, "Scaled", transform=axs[row, col + 1].transAxes, ha='center',
                                       va='bottom', fontsize=6)

            # Remove any unused subplots
            for i in range(plots_on_page, rows_on_page * (cols // 2)):
                row = i // (cols // 2)
                col = (i % (cols // 2)) * 2
                fig.delaxes(axs[row, col])
                fig.delaxes(axs[row, col + 1])

            pdf.savefig(fig)
            plt.close(fig)

    print(f"PDF saved as {output_pdf_name}")


if __name__ == '__main__':
    min_score = 0.7
    min_matched_peak = 4
    min_spectral_usage = 0.05

    # file_name = '/Users/shipei/Documents/projects/ms1_id/bin/msi/analysis/mouse_brain/all_brain_annotations.tsv'
    # plot_multiple_figures(file_name, 'all_brain_mirror_plot.pdf', min_score, min_matched_peak, min_spectral_usage)

    file_name = '/bin/ms1id/msi/analysis/mouse_brain/all_brain_annotations_verified.tsv'
    plot_multiple_figures(file_name, 'all_brain_verified_sqrt_mirror_plot.pdf',
                          min_score,
                          min_matched_peak,
                          min_spectral_usage,
                          peak_int_power=0.5)
    plot_multiple_figures(file_name, 'all_brain_verified_mirror_plot.pdf',
                          min_score,
                          min_matched_peak,
                          min_spectral_usage,
                          peak_int_power=1)

    # file_name = '/Users/shipei/Documents/projects/ms1_id/imaging/mouse_body/wb xenograft in situ metabolomics test - rms_corrected/ms1_id_annotations_derep.tsv'
    # plot_multiple_figures(file_name, 'mouse_body_mirror_plot.pdf', min_score, min_matched_peak, min_spectral_usage)
    #
    # file_name = '/Users/shipei/Documents/projects/ms1_id/imaging/mouse_kidney/mouse kidney - root mean square - metaspace/ms1_id_annotations_derep.tsv'
    # plot_multiple_figures(file_name, 'mouse_kidney_mirror_plot.pdf', min_score, min_matched_peak, min_spectral_usage)
    #
    # file_name = '/Users/shipei/Documents/projects/ms1_id/imaging/mouse_liver/Mouse liver_DMAN_200x200_25um/ms1_id_annotations_derep.tsv'
    # plot_multiple_figures(file_name, 'mouse_liver_mirror_plot.pdf', min_score, min_matched_peak, min_spectral_usage)



import os

import pandas as pd


def write_ms1_id_results(ms1_spec_ls, save=True, save_dir=None):
    """
    Output the annotated ms1 spectra
    :param ms1_spec_ls: a list of PseudoMS2-like object
    :param save: bool, whether to save the results
    :param save_dir: str, path to save the results
    :return: None
    """

    # only write out spectra with annotations
    ms1_spec_ls = [spec for spec in ms1_spec_ls if spec.annotated]

    out_list = []
    for spec in ms1_spec_ls:
        # pseudo_ms2_str: mz1 int1; mz2 int2; ...
        pseudo_ms2_str = ' '.join([f"{mz:.4f} {intensity:.0f};" for mz, intensity in zip(spec.mzs, spec.intensities)])

        for annotation in spec.annotation_ls:
            out_list.append({
                'name': annotation.name,
                'mz': round(annotation.mz, 4),
                'matched_score': round(annotation.score, 4),
                'matched_peak': annotation.matched_peak,
                'spectral_usage': round(annotation.spectral_usage, 4) if annotation.spectral_usage else None,
                # 'search_eng_matched_id': annotation.search_eng_matched_id,
                'precursor_mz': round(annotation.precursor_mz, 4),
                'precursor_type': annotation.precursor_type,
                'formula': annotation.formula,
                'inchikey': annotation.inchikey,
                # 'instrument_type': annotation.instrument_type,
                'collision_energy': annotation.collision_energy,
                'db_name': annotation.db_name,
                'db_id': annotation.db_id,
                'pseudo_ms2': pseudo_ms2_str,
            })

    out_df = pd.DataFrame(out_list)

    if save and save_dir:
        save_path = os.path.join(save_dir, 'ms1_id_annotations_all.tsv')
        out_df.to_csv(save_path, index=False, sep='\t')

        # save a dereplicated version
        if 'matched_score' in out_df.columns:
            # sort
            out_df = out_df.sort_values(['spectral_usage', 'matched_score', 'matched_peak'],
                                        ascending=[False, False, False])

            # dereplicate by [inchikey, rounded precursor mz]  # rounded precursor mz indicating precursor type
            out_df['rounded_precursor_mz'] = out_df['precursor_mz'].round(1)
            out_df = out_df.drop_duplicates(['inchikey', 'rounded_precursor_mz'], keep='first')
            out_df.drop(columns=['rounded_precursor_mz'], inplace=True)

            out_df.to_csv(save_path.replace('_all.tsv', '_derep.tsv'), index=False, sep='\t')

    return out_df


def write_pseudoms2_to_mgf(pseudo_ms2_spectra, mgf_path):
    """
    Write pseudo MS2 spectra to an MGF file
    :param pseudo_ms2_spectra: list of PseudoMS2 objects
    :param mgf_path: str, path to save the MGF file
    :return: None
    """
    with open(mgf_path, 'w') as f:
        idx = 1
        for spec in pseudo_ms2_spectra:
            f.write(f"BEGIN IONS\n")
            f.write(f"PEPMASS={round(spec.t_mz, 5)}\n")
            f.write(f"SCANS={idx}\n")
            for mz, intensity in zip(spec.mzs, spec.intensities):
                f.write(f"{mz:.5f} {intensity:.0f}\n")
            f.write(f"END IONS\n\n")
            idx += 1
    return None


import numpy as np
from numba import njit


@njit
def _centroid_spectrum(peaks, centroid_mode='max', peak_height_requirement=True,
                       width_da=0.005, width_ppm=25.0) -> np.ndarray:
    """Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da,
    only merging peaks with lower intensity than the target peak."""
    # Construct a new spectrum to avoid memory reallocation.
    peaks_new = np.zeros((peaks.shape[0], 2), dtype=np.float32)
    peaks_new_len = 0

    # Get the intensity argsort order.
    intensity_order = np.argsort(peaks[:, 1])

    # Iterate through the peaks from high to low intensity.
    for idx in intensity_order[::-1]:
        if peaks[idx, 1] > 0:
            mz_delta_allowed_ppm_in_da = peaks[idx, 0] * width_ppm * 1e-6
            mz_delta_allowed = max(width_da, mz_delta_allowed_ppm_in_da)

            # Find indices of peaks to merge.
            if peak_height_requirement:
                indices = np.where(np.logical_and(np.abs(peaks[:, 0] - peaks[idx, 0]) <= mz_delta_allowed,
                                                  peaks[:, 1] < peaks[idx, 1]))[0]
            else:
                indices = np.where(np.abs(peaks[:, 0] - peaks[idx, 0]) <= mz_delta_allowed)[0]

            if centroid_mode == 'max':
                # Merge the peaks
                peaks_new[peaks_new_len, 0] = peaks[idx, 0]
                peaks_new[peaks_new_len, 1] = peaks[idx, 1]
                peaks_new_len += 1

                peaks[indices, 1] = 0

            elif centroid_mode == 'sum':
                # Merge the peaks
                intensity_sum = peaks[idx, 1]
                intensity_weighted_mz_sum = peaks[idx, 1] * peaks[idx, 0]
                for i in indices:
                    intensity_sum += peaks[i, 1]
                    intensity_weighted_mz_sum += peaks[i, 1] * peaks[i, 0]
                    peaks[i, 1] = 0  # Set the intensity of the merged peaks to 0

                peaks_new[peaks_new_len, 0] = intensity_weighted_mz_sum / intensity_sum
                peaks_new[peaks_new_len, 1] = intensity_sum
                peaks_new_len += 1

                # Set the intensity of the target peak to 0
                peaks[idx, 1] = 0

    # Return the new spectrum.
    peaks_new = peaks_new[:peaks_new_len]
    return peaks_new[np.argsort(peaks_new[:, 0])]


def check_centroid_for_search(peaks, width_da=0.105):
    """Check if the spectrum is centroided. True for centroided and False for not centroided."""

    # Check if the spectrum is centroided
    return True if np.all(np.diff(peaks[:, 0]) >= width_da) else False


def centroid_spectrum_for_search(peaks, width_da=0.105, width_ppm=25.0):
    """Centroid a spectrum for search."""
    if len(peaks) == 0:
        return peaks

    if check_centroid_for_search(peaks, width_da=width_da):
        return peaks

    # Sort the peaks by m/z.
    peaks = peaks[np.argsort(peaks[:, 0])]

    # centroid the spectrum
    peaks = _centroid_spectrum(peaks, centroid_mode='max', width_da=width_da, width_ppm=width_ppm)
    return peaks


def consensus_spectrum(peaks, width_da=0.025, width_ppm=25.0):
    """Centroid a spectrum for search."""
    if len(peaks) == 0:
        return peaks

    if check_centroid_for_search(peaks, width_da=width_da):
        return peaks

    # Sort the peaks by m/z.
    peaks = peaks[np.argsort(peaks[:, 0])]

    # centroid the spectrum
    peaks = _centroid_spectrum(peaks, centroid_mode='sum', peak_height_requirement=False,
                               width_da=width_da, width_ppm=width_ppm)
    return peaks



import matplotlib.pyplot as plt
from pyteomics import mzml
import numpy as np
import pickle
import pandas as pd
from flash_cos import FlashCos
from molmass import Formula


_adduct_pos = [
    {'name': '[M+H]+', 'm': 1, 'charge': 1, 'mass': 1.00727645223},
    {'name': '[M+Na]+', 'm': 1, 'charge': 1, 'mass': 22.989220702},
    {'name': '[M+K]+', 'm': 1, 'charge': 1, 'mass': 38.9631579064},
    {'name': '[M+NH4]+', 'm': 1, 'charge': 1, 'mass': 18.03382555335},
    {'name': '[M+H-H2O]+', 'm': 1, 'charge': 1, 'mass': -17.0032882318},
    {'name': '[2M+H]+', 'm': 2, 'charge': 1, 'mass': 1.00727645223},
    {'name': '[2M+Na]+', 'm': 2, 'charge': 1, 'mass': 22.989220702},
    {'name': '[2M+K]+', 'm': 2, 'charge': 1, 'mass': 38.9631579064},
    {'name': '[2M+NH4]+', 'm': 2, 'charge': 1, 'mass': 18.03382555335},
    {'name': '[M+2H]2+', 'm': 1, 'charge': 2, 'mass': 2.01455290446},
]


def make_library_main(peak_scale_k=8):
    db_ls = []
    db_ls1 = make_library('/data/std_mix/data_2/mixture_Drugs_DDA.mzML',
                          '/data/std_mix/drug_std.tsv')
    print('Drugs:', len(db_ls1))
    db_ls2 = make_library('/data/std_mix/data_2/mixture_Bile_acids_DDA.mzML',
                          '/data/std_mix/BA_std.tsv')
    print('BA:', len(db_ls2))
    db_ls3 = make_library('/data/std_mix/data_2/sixmix_2_DDA.mzML',
                          '/data/std_mix/six_std.tsv')
    print('Six:', len(db_ls3))

    db_ls.extend(db_ls1)
    db_ls.extend(db_ls2)
    db_ls.extend(db_ls3)

    if peak_scale_k is not None:
        out_name = f'/Users/shipei/Documents/projects/ms1_id/data/std_mix/std_k{peak_scale_k}.pkl'
    else:
        out_name = '/Users/shipei/Documents/projects/ms1_id/data/std_mix/std.pkl'

    index_library(db_ls, out_name, peak_scale_k)


def make_library(mzml_file, tsv_file):
    db_ls = []

    df = pd.read_csv(tsv_file, sep='\t', low_memory=False)
    df['neutral_mass'] = df['formula'].apply(calc_exact_mass)

    with mzml.read(mzml_file) as reader:
        spectra = list(reader)

    for i, row in df.iterrows():
        for adduct in _adduct_pos:
            adduct_mz = (row['neutral_mass'] * adduct['m'] + adduct['mass']) / adduct['charge']
            spectrum = find_closest_ms2(spectra, adduct_mz, row['rt'], rt_tolerance=0.05, mass_tolerance=0.01)

            if spectrum is not None:
                peaks = np.column_stack((spectrum['m/z array'], spectrum['intensity array']))

                db_ls.append({
                    'id': f"{i}_{adduct['name']}",
                    'name': row['name'],
                    'precursor_mz': adduct_mz,
                    'formula': row['formula'],
                    'precursor_type': adduct['name'],
                    'ion_mode': 'positive',
                    'original_peaks': peaks,
                    'peaks': peaks
                })

    return db_ls


def index_library(db_ls, out_name, peak_scale_k):
    mz_tol = 0.05

    search_engine = FlashCos(max_ms2_tolerance_in_da=mz_tol * 1.005,
                             mz_index_step=0.0001,
                             peak_scale_k=peak_scale_k,
                             peak_intensity_power=0.5)
    print('building index')
    search_engine.build_index(db_ls,
                              max_indexed_mz=1000,
                              precursor_ions_removal_da=0.5,
                              noise_threshold=0.0,
                              min_ms2_difference_in_da=mz_tol * 2.02,
                              clean_spectra=True)

    # save as pickle
    with open(out_name, 'wb') as file:
        # Dump the data into the file
        pickle.dump(search_engine, file)


def calc_exact_mass(formula):
    """
    Calculate the exact mass for a given formula string
    """
    try:
        f = Formula(formula)
        return f.monoisotopic_mass
    except:
        return None


def find_closest_ms2(spectra, target_mz, target_rt, rt_tolerance=0.05, mass_tolerance=0.01):
    closest_spectrum = None
    min_rt_diff = float('inf')

    for spectrum in spectra:
        if spectrum['ms level'] != 2:
            continue

        rt = spectrum.get('scanList', {}).get('scan', [{}])[0].get('scan start time', 0)
        if abs(rt - target_rt) > rt_tolerance:
            continue

        precursor_mz = \
        spectrum.get('precursorList', {}).get('precursor', [{}])[0].get('selectedIonList', {}).get('selectedIon', [{}])[
            0].get('selected ion m/z')
        if precursor_mz is None or abs(precursor_mz - target_mz) > mass_tolerance:
            continue

        rt_diff = abs(rt - target_rt)
        if rt_diff < min_rt_diff:
            closest_spectrum = spectrum
            min_rt_diff = rt_diff

    return closest_spectrum


def get_spectrum_by_scan(mzml_file, scan_number):
    """
    Function to get a spectrum by scan number from an mzML file.
    :param mzml_file: Path to the mzML file.
    :param scan_number: The scan number of the spectrum to retrieve.
    :return: A dictionary containing the m/z array, intensity array, and metadata of the spectrum.
    """
    with mzml.read(mzml_file) as reader:
        for spectrum in reader:
            if spectrum['index'] + 1 == scan_number:  # mzML scan numbers are 1-based, but index is 0-based
                return {
                    'mz': np.array(spectrum['m/z array']),
                    'intensity': np.array(spectrum['intensity array'])
                }

    return None


def plot_spectrum(spectrum):
    """
    Function to plot a spectrum.
    :param spectrum: A dictionary containing 'm/z array' and 'intensity array'.
    """
    plt.figure(figsize=(10, 6))
    plt.vlines(spectrum['mz'], 0, spectrum['intensity'])
    plt.xlabel('m/z')
    plt.ylabel('Intensity')
    plt.show()


if __name__ == '__main__':
    make_library_main(peak_scale_k=0)
    # make_library_main(peak_scale_k=3)
    # make_library_main(peak_scale_k=5)
    # make_library_main(peak_scale_k=8)
    # make_library_main(peak_scale_k=9)
    make_library_main(peak_scale_k=10)
    # make_library_main(peak_scale_k=13)
    # make_library_main(peak_scale_k=15)
    # make_library_main(peak_scale_k=18)




"""
This module provides functions to preprocess MS2 spectra.
Some functions are adapted from ms_entropy package.
"""

import numpy as np


def preprocess_ms2(peaks, prec_mz,
                   min_mz=-1, max_mz=-1,
                   relative_intensity_cutoff=0.01,
                   min_ms2_difference_in_da=0.05,
                   min_ms2_difference_in_ppm=-1,
                   top6_every_50da=False,
                   peak_scale_k=None,
                   peak_intensity_power=1,
                   peak_norm=None) -> np.ndarray:
    """
    Main function to preprocess MS2 spectra.

        Clean, centroid, and normalize a spectrum with the following steps:
        1. Remove empty peaks (m/z <= 0 or intensity <= 0).
        2. Remove peaks with m/z >= max_mz or m/z < min_mz.
        3. Centroid the spectrum by merging peaks within min_ms2_difference_in_da.
        4. Remove peaks with intensity < relative_intensity_cutoff * max_intensity.
        5. Keep only the top max_peak_num peaks within every 50 Da.
        6. Transform the peak intensity.
        7. peak_intensity_power is used to transform the intensity to intensity^peak_intensity_power.
        8. Normalize the intensity.

        The cleaned spectrum will be sorted by m/z in ascending order.

    Parameters
    ----------
    peaks : np.ndarray in shape (n_peaks, 2)
        A 2D array of shape (n_peaks, 2) where the first column is m/z and the second column is intensity.

    prec_mz : float
        The precursor m/z value.

    min_mz : float, optional
        The minimum m/z to keep. Defaults to -1, which will skip removing peaks with m/z < min_mz.

    max_mz : float, optional
        The maximum m/z to keep. Defaults to -1, which will skip removing peaks with m/z >= max_mz.

    relative_intensity_cutoff : float, optional
        The minimum intensity to keep. Defaults to 0.01, which will remove peaks with intensity < 0.01 * max_intensity.

    min_ms2_difference_in_da : float, optional
        The minimum m/z difference between two peaks in the resulting spectrum. Defaults to 0.05, which will merge peaks within 0.05 Da. If a negative value is given, the min_ms2_difference_in_ppm will be used instead.

    min_ms2_difference_in_ppm : float, optional
        The minimum m/z difference between two peaks in the resulting spectrum. Defaults to -1, which will use the min_ms2_difference_in_da instead. If a negative value is given, the min_ms2_difference_in_da will be used instead.

    top6_every_50da : bool, optional
        Whether to keep only the top 6 peaks within every 50 Da. Defaults to False, which will keep all peaks.

    peak_intensity_power : float, optional
        The power to transform the peak intensity. Defaults to 1, which will not transform the intensity.

    peak_norm : str, optional
        The peak normalization method. The available methods are: 'sum', 'sum_sq'.
    """

    # Check the input spectrum and convert it to numpy array with shape (n, 2) and dtype np.float32.
    peaks = np.asarray(peaks, dtype=np.float32, order="C")
    if len(peaks) == 0:
        return np.zeros((0, 2), dtype=np.float32)
    assert peaks.ndim == 2 and peaks.shape[1] == 2, "The input spectrum must be a numpy array with shape (n, 2)."

    # Step 1. Remove empty peaks (m/z <= 0 or intensity <= 0).
    peaks = peaks[np.bitwise_and(peaks[:, 0] > 0, peaks[:, 1] > 0)]

    # Step 2. Remove peaks with m/z >= max_mz or m/z < min_mz.
    if min_mz is not None and min_mz > 0:
        peaks = peaks[peaks[:, 0] > min_mz]

    if max_mz is not None and max_mz > 0:
        peaks = peaks[peaks[:, 0] < max_mz]

    if peaks.shape[0] == 0:
        return np.zeros((0, 2), dtype=np.float32)

    # Step 3. Centroid the spectrum by merging peaks within min_ms2_difference_in_da.
    # The peaks will be sorted by m/z in ascending order.
    if min_ms2_difference_in_ppm > 0:
        peaks = centroid_spectrum(peaks, ms2_da=-1, ms2_ppm=min_ms2_difference_in_ppm)
    elif min_ms2_difference_in_da > 0:
        peaks = centroid_spectrum(peaks, ms2_da=min_ms2_difference_in_da, ms2_ppm=-1)
    else:
        pass

    # Step 4. Remove peaks with intensity < relative_intensity_cutoff * max_intensity.
    if relative_intensity_cutoff is not None:
        peaks = peaks[peaks[:, 1] >= relative_intensity_cutoff * np.max(peaks[:, 1])]

    # Step 5. Keep only the top 6 peaks within every 50 Da
    if top6_every_50da:
        peaks = top_n_per_mz_range(peaks, n_peaks=6, mz_range=50)

    # Step 6. scale the intensity.
    if peak_scale_k is not None:
        _prec_mz = prec_mz if prec_mz > 0 else np.max(peaks[:, 0])
        scaling_factor = peaks[:, 0] / _prec_mz * peak_scale_k
        peaks[:, 1] = peaks[:, 1] * np.exp(scaling_factor)

    # Step 7. Power the intensity.
    peaks[:, 1] = np.power(peaks[:, 1], peak_intensity_power)

    # Step 8. Normalize the intensity.
    if peak_norm is not None:
        if peak_norm == 'sum':
            peaks[:, 1] = peaks[:, 1] / np.sum(peaks[:, 1])
        elif peak_norm == 'sum_sq':
            peaks[:, 1] = peaks[:, 1] / np.sqrt(np.sum(peaks[:, 1] ** 2))

    return peaks


def top_n_per_mz_range(peaks: np.ndarray, n_peaks: int = 6, mz_range: int = 50) -> np.ndarray:
    """
    Keep only the top n_peaks peaks within every mz_range Da.
    """

    new_peaks = []  # List to store the filtered peaks
    # Calculate the range of m/z values
    min_mz_val, max_mz_val = np.min(peaks[:, 0]), np.max(peaks[:, 0])

    # Iterate over segments of 50 Da, starting from the minimum m/z value
    mz_start = 0  # Starting from 0 as mentioned
    while mz_start <= max_mz_val:
        mz_end = mz_start + mz_range
        # Filter peaks within the current m/z range
        segment_peaks = peaks[np.logical_and(peaks[:, 0] >= mz_start, peaks[:, 0] < mz_end)]

        # Check if there are any peaks in the segment
        if len(segment_peaks) > 0:
            # Sort the peaks by intensity in descending order to get the top peaks
            sorted_peaks = segment_peaks[np.argsort(-segment_peaks[:, 1])]
            # Keep only the top n_peaks, or all if there are fewer than n_peaks
            top_peaks = sorted_peaks[:min(n_peaks, len(sorted_peaks))]
            # Append the top peaks to the new_peaks list
            new_peaks.extend(top_peaks)

        mz_start += mz_range  # Move to the next segment

    # Convert the list back to a numpy array
    peaks = np.array(new_peaks)
    peaks = peaks[np.argsort(peaks[:, 0])]

    return peaks


def centroid_spectrum(peaks: np.ndarray, ms2_da: float = -1, ms2_ppm: float = -1) -> np.ndarray:
    """
    Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da.

    Parameters
    ----------
    peaks : np.ndarray in shape (n_peaks, 2), dtype=np.float32
        A 2D array of shape (n_peaks, 2) where the first column is m/z and the second column is intensity.

    ms2_ppm : float, optional
        The m/z tolerance in ppm. Defaults to -1, which will use ms2_da instead.

    ms2_da : float, optional
        The m/z tolerance in Da. Defaults to -1, which will use ms2_ppm instead.

        **Note: ms2_ppm and ms2_da cannot be both negative, if so, an Exception will be raised. If both are positive, ms2_da will be used.**

    Returns
    -------
    np.ndarray in shape (n_peaks, 2), dtype=np.float32
        The resulting spectrum after centroiding have been sorted by m/z, and the difference between two peaks will be >= ms2_ppm or >= ms2_da.
    """
    peaks = np.asarray(peaks, dtype=np.float32, order="C")
    # Sort the peaks by m/z.
    peaks = peaks[np.argsort(peaks[:, 0])]
    is_centroided: int = _check_centroid(peaks, ms2_da=ms2_da, ms2_ppm=ms2_ppm)
    while is_centroided == 0:
        peaks = _centroid_spectrum(peaks, ms2_da=ms2_da, ms2_ppm=ms2_ppm)
        is_centroided = _check_centroid(peaks, ms2_da=ms2_da, ms2_ppm=ms2_ppm)
    return peaks


def _centroid_spectrum(peaks: np.ndarray, ms2_da: float = -1, ms2_ppm: float = -1) -> np.ndarray:
    """Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da."""
    # Construct a new spectrum to avoid memory reallocation.
    peaks_new = np.zeros((peaks.shape[0], 2), dtype=np.float32)
    peaks_new_len = 0

    # Get the intensity argsort order.
    intensity_order = np.argsort(peaks[:, 1])

    mz_delta_allowed_left = ms2_da
    mz_delta_allowed_right = ms2_da
    # Iterate through the peaks from high to low intensity.
    for idx in intensity_order[::-1]:
        if ms2_ppm > 0:
            mz_delta_allowed_left = peaks[idx, 0] * ms2_ppm * 1e-6
            # For the right boundary, the mz_delta_allowed_right = peaks[right_idx, 0] * ms2_ppm * 1e-6 = peaks[idx, 0] / (1 - ms2_ppm * 1e-6)
            mz_delta_allowed_right = peaks[idx, 0] / (1 - ms2_ppm * 1e-6)

        if peaks[idx, 1] > 0:
            # Find the left boundary.
            left_idx = idx - 1
            while left_idx >= 0 and peaks[idx, 0] - peaks[left_idx, 0] <= mz_delta_allowed_left:
                left_idx -= 1
            left_idx += 1

            # Find the right boundary.
            right_idx = idx + 1
            while right_idx < peaks.shape[0] and peaks[right_idx, 0] - peaks[idx, 0] <= mz_delta_allowed_right:
                right_idx += 1

            # Merge the peaks within the boundary. The new m/z is the weighted average of the m/z and intensity.
            intensity_sum = np.sum(peaks[left_idx:right_idx, 1])
            intensity_weighted_mz_sum = np.sum(peaks[left_idx:right_idx, 1] * peaks[left_idx:right_idx, 0])
            peaks_new[peaks_new_len, 0] = intensity_weighted_mz_sum / intensity_sum
            peaks_new[peaks_new_len, 1] = intensity_sum
            peaks_new_len += 1

            # Set the intensity of the merged peaks to 0.
            peaks[left_idx:right_idx, 1] = 0

    # Return the new spectrum.
    peaks_new = peaks_new[:peaks_new_len]
    return peaks_new[np.argsort(peaks_new[:, 0])]


def _check_centroid(peaks: np.ndarray, ms2_da: float = -1, ms2_ppm: float = -1) -> int:
    """Check if the spectrum is centroided. 0 for False and 1 for True."""
    if peaks.shape[0] <= 1:
        return 1

    if ms2_ppm >= 0:
        # Use ms2_ppm to check if the spectrum is centroided.
        return 1 if np.all(np.diff(peaks[:, 0]) >= peaks[1:, 0] * ms2_ppm * 1e-6) else 0
    elif ms2_da >= 0:
        # Use ms2_da to check if the spectrum is centroided.
        return 1 if np.all(np.diff(peaks[:, 0]) >= ms2_da) else 0



# A module to align metabolic features from different samples
# Isotopes and in-source fragments are not considered in the alignment

# Import modules
import numpy as np
import pandas as pd
import os
from tqdm import tqdm
from scipy.interpolate import interp1d
from scipy.stats import zscore
import pickle

from ms1_id.lcms.masscube_raw_data_utils import read_raw_file_to_obj


def feature_alignment(path, parameters, drop_by_fill_pct_ratio=0.1):
    """
    A function to align the features from individual files (.txt).

    Parameters
    ----------------------------------------------------------
    path: str
        The path to the feature tables.
    parameters: Params object
        The parameters for alignment.

    Returns
    ----------------------------------------------------------
    feature_table: DataFrame
        The aligned feature table.
    """

    # STEP 1: get names of individual files (.txt)
    txt_file_names = parameters.sample_names
    txt_file_names = [f + ".txt" for f in txt_file_names]
    txt_file_names = [os.path.join(path, name) for name in txt_file_names]

    # STEP 2: initiate aligned features and parameters
    features = []
    mz_tol = parameters.align_mz_tol
    rt_tol = parameters.align_rt_tol
    file_quality_arr = np.ones(len(parameters.sample_names), dtype=bool)
    # select anchors for retention time correction
    if parameters.run_rt_correction:
        mz_ref, rt_ref = rt_anchor_selection(txt_file_names[:20])
        rt_cor_functions = {}

    # STEP 3: read individual feature tables and align features
    for i, file_name in enumerate(tqdm(parameters.sample_names)):
        # check if the file exists, if not, add the file to problematic_files
        if not os.path.exists(txt_file_names[i]):
            file_quality_arr[i] = False
            continue

        # read feature table
        current_table = pd.read_csv(txt_file_names[i], low_memory=False, sep="\t")
        current_table = current_table[
            current_table["MS2"].notna() | (current_table["length"] >= parameters.min_scan_num_for_alignment)]
        # sort current table by peak height from high to low
        current_table = current_table.sort_values(by="peak_height", ascending=False)
        current_table.index = range(len(current_table))
        avail_roi = np.ones(len(current_table), dtype=bool)
        # retention time correction
        if parameters.run_rt_correction and parameters.individual_sample_groups[i] != 'blank':
            rt_arr = current_table["RT"].values
            rt_max = np.max(rt_arr)
            rt_arr, model = retention_time_correction(mz_ref, rt_ref, current_table["m/z"].values, rt_arr,
                                                      rt_max=rt_max, return_model=True)
            current_table["RT"] = rt_arr
            rt_cor_functions[file_name] = model

        if len(features) > 0:
            for f in features:
                # search mz and rt in the current feature table
                v = np.logical_and(np.abs(f.mz - current_table["m/z"].values) < mz_tol,
                                   np.abs(f.rt - current_table["RT"].values) < rt_tol)
                v = np.logical_and(v, avail_roi)
                v = np.where(v)[0]

                if len(v) > 0:
                    v = v[0]
                    f.mz_seq[i] = current_table.loc[v, "m/z"]
                    f.rt_seq[i] = current_table.loc[v, "RT"]
                    f.peak_height_seq[i] = current_table.loc[v, "peak_height"]
                    f.peak_area_seq[i] = current_table.loc[v, "peak_area"]
                    f.ms2_seq.append(current_table.loc[v, "MS2"])
                    f.detected_seq[i] = True
                    f.roi_id_seq[i] = current_table.loc[v, "ID"]
                    avail_roi[v] = False
                    # check if this file can be the reference file
                    if current_table.loc[v, "peak_height"] > f.highest_intensity:
                        f.highest_intensity = current_table.loc[v, "peak_height"]
                        f.gaussian_similarity = current_table.loc[v, "Gaussian_similarity"]
                        f.noise_level = current_table.loc[v, "noise_level"]
                        f.asymmetry_factor = current_table.loc[v, "asymmetry_factor"]
                        f.charge_state = current_table.loc[v, "charge"]
                        f.is_isotope = current_table.loc[v, "is_isotope"]
                        f.isotopes = current_table.loc[v, "isotopes"]
                        f.is_in_source_fragment = current_table.loc[v, "is_in_source_fragment"]
                        f.adduct_type = current_table.loc[v, "adduct"]
                        f.reference_file = file_name

        # if an ROI is not detected in the previous files, add it to the features
        if np.sum(avail_roi) > 0:
            v = np.where(avail_roi)[0]
            for j in v:
                f = Feature(file_number=len(parameters.sample_names))
                f.mz = current_table.loc[j, "m/z"]
                f.rt = current_table.loc[j, "RT"]
                f.mz_seq[i] = current_table.loc[j, "m/z"]
                f.rt_seq[i] = current_table.loc[j, "RT"]
                f.peak_height_seq[i] = current_table.loc[j, "peak_height"]
                f.peak_area_seq[i] = current_table.loc[j, "peak_area"]
                f.ms2_seq.append(current_table.loc[j, "MS2"])
                f.detected_seq[i] = True
                f.roi_id_seq[i] = current_table.loc[j, "ID"]
                f.highest_intensity = current_table.loc[j, "peak_height"]
                f.gaussian_similarity = current_table.loc[j, "Gaussian_similarity"]
                f.noise_level = current_table.loc[j, "noise_level"]
                f.asymmetry_factor = current_table.loc[j, "asymmetry_factor"]
                f.charge_state = current_table.loc[j, "charge"]
                f.is_isotope = current_table.loc[j, "is_isotope"]
                f.isotopes = current_table.loc[j, "isotopes"]
                f.is_in_source_fragment = current_table.loc[j, "is_in_source_fragment"]
                f.adduct_type = current_table.loc[j, "adduct"]
                f.reference_file = file_name
                features.append(f)

        # summarize (calculate the average mz and rt) and reorder the features
        features = sorted(features, key=lambda x: x.highest_intensity, reverse=True)
        for f in features:
            f.calculate_mzrt()

    # STEP 4: drop the problematic files and index the features
    if np.sum(file_quality_arr) > 0:
        parameters.sample_names = [parameters.sample_names[i] for i in range(len(parameters.sample_names)) if
                                   file_quality_arr[i]]
        parameters.individual_sample_groups = [parameters.individual_sample_groups[i] for i in
                                               range(len(parameters.individual_sample_groups)) if file_quality_arr[i]]
        for f in features:
            f.mz_seq = f.mz_seq[file_quality_arr]
            f.rt_seq = f.rt_seq[file_quality_arr]
            f.peak_height_seq = f.peak_height_seq[file_quality_arr]
            f.peak_area_seq = f.peak_area_seq[file_quality_arr]
            f.detected_seq = f.detected_seq[file_quality_arr]
            # remove nan in f.ms2_seq
            f.ms2_seq = [ms2 for ms2 in f.ms2_seq if ms2 == ms2]

    # STEP 5: calculate the fill percentage and remove the features with fill percentage less than 0.1
    blank_num = len([x for x in parameters.individual_sample_groups if 'blank' in x])
    if blank_num > 0:
        features = [f for f in features if np.sum(f.detected_seq[:-blank_num]) > drop_by_fill_pct_ratio * (
                    len(parameters.sample_names) - blank_num)]
    else:
        features = [f for f in features if
                    np.sum(f.detected_seq) > drop_by_fill_pct_ratio * len(parameters.sample_names)]

    for i in range(len(features)):
        features[i].id = i

    # output the models to pickle files
    if parameters.run_rt_correction:
        with open(os.path.join(parameters.project_dir, "rt_correction_models.pkl"), "wb") as f:
            pickle.dump(rt_cor_functions, f)

    return features


def gap_filling(features, parameters, mode='forced_peak_picking'):
    """
    A function to fill the gaps in the aligned feature table.

    Parameters
    ----------------------------------------------------------
    features: list
        The aligned features.
    parameters: Params object
        The parameters for gap filling.
    mode: str
        The mode for gap filling.
        'forced_peak_picking': fill the gaps by forced peak picking.
        '0.1_min_intensity': fill the gaps by the minimum intensity * 0.1 (no available yet)

    Returns
    ----------------------------------------------------------
    features: list
        The aligned features with filled gaps.
    """

    # fill the gaps by forced peak picking
    if mode == 'forced_peak_picking':
        raw_file_names = os.listdir(parameters.sample_dir)
        raw_file_names = [f for f in raw_file_names if f.lower().endswith(".mzml") or f.lower().endswith(".mzxml")]
        raw_file_names = [f for f in raw_file_names if not f.startswith(".")]
        raw_file_names = [os.path.join(parameters.sample_dir, f) for f in raw_file_names]

        # if retention time correction is applied, read the model
        if parameters.run_rt_correction and os.path.exists(
                os.path.join(parameters.project_dir, "rt_correction_models.pkl")):
            with open(os.path.join(parameters.project_dir, "rt_correction_models.pkl"), "rb") as f:
                rt_cor_functions = pickle.load(f)
        else:
            rt_cor_functions = None

        for i, file_name in enumerate(tqdm(parameters.sample_names)):
            matched_raw_file_name = [f for f in raw_file_names if file_name in f]
            if len(matched_raw_file_name) == 0:
                continue
            else:
                matched_raw_file_name = matched_raw_file_name[0]
                d = read_raw_file_to_obj(matched_raw_file_name, int_tol=parameters.int_tol, read_ms2=False)

                # correct retention time if model is available
                if rt_cor_functions is not None and file_name in rt_cor_functions.keys():
                    f = rt_cor_functions[file_name]
                    if f is not None:
                        d.correct_retention_time(f)

                for f in features:
                    if not f.detected_seq[i]:
                        _, eic_int, _, _ = d.get_eic_data(f.mz, f.rt, parameters.align_mz_tol, 0.05)
                        if len(eic_int) > 0:
                            f.peak_height_seq[i] = np.max(eic_int)

    # calculate the fill percentage after gap filling (blank samples are not included)
    blank_num = len([x for x in parameters.individual_sample_groups if 'blank' in x])
    for f in features:
        f.fill_percentage = (np.sum(f.detected_seq) - blank_num) / (len(parameters.sample_names) - blank_num) * 100
    return features


def output_feature_table(feature_table, output_path):
    """
    A function to output the aligned feature table.

    Parameters
    ----------------------------------------------------------
    feature_table: DataFrame
        The aligned feature table.
    output_path: str
        The path to save the aligned feature table.
    """

    # keep four digits for the m/z column and three digits for the RT column
    feature_table["m/z"] = feature_table["m/z"].apply(lambda x: round(x, 4))
    feature_table["RT"] = feature_table["RT"].apply(lambda x: round(x, 3))
    feature_table['fill_percentage'] = feature_table['fill_percentage'].apply(lambda x: round(x, 2))
    feature_table['similarity'] = feature_table['similarity'].astype(float)
    feature_table['similarity'] = feature_table['similarity'].apply(lambda x: round(x, 4))

    feature_table.to_csv(output_path, index=False, sep="\t")


def retention_time_correction(mz_ref, rt_ref, mz_arr, rt_arr, rt_max=50, mode='linear_interpolation', mz_tol=0.015,
                              rt_tol=2.0, found_marker_ratio=0.4, return_model=False):
    """
    To correct retention times for feature alignment.

    There are three steps:
    1. Find the selected anchors in the given data.
    2. Create a model to correct retention times.
    3. Correct retention times.

    Parameters
    ----------
    mz_ref: np.array
        The m/z values of the reference features.
    rt_ref: np.array
        The retention times of the reference features.
    mz_arr: np.array
        The m/z values of the features to be corrected.
    rt_arr: np.array
        The retention times of the features to be corrected.
    mode: str
        The mode for retention time correction.
        'linear_interpolation': linear interpolation for retention time correction.
    mz_tol: float
        The m/z tolerance for selecting anchors.
    rt_tol: float
        The retention time tolerance for selecting anchors.

    Returns
    -------
    rt_corr: np.array
        The corrected retention times.
    """

    mz_matched = []
    rt_matched = []
    idx_matched = []

    for i in range(len(mz_ref)):
        v = np.logical_and(np.abs(mz_arr - mz_ref[i]) < mz_tol, np.abs(rt_arr - rt_ref[i]) < rt_tol)
        v = np.where(v)[0]
        if len(v) > 0:
            mz_matched.append(mz_arr[v[0]])
            rt_matched.append(rt_arr[v[0]])
            idx_matched.append(i)
    rt_ref = rt_ref[idx_matched]

    if len(idx_matched) < found_marker_ratio * len(mz_ref):
        if return_model:
            return rt_arr, None
        else:
            return rt_arr

    # remove outliers
    v = rt_ref - np.array(rt_matched)
    z = zscore(v)
    outliers = np.where(np.logical_and(np.abs(z) > 1, np.abs(v) > 0.1))[0]
    if len(outliers) > 0:
        rt_ref = np.delete(rt_ref, outliers)
        rt_matched = np.delete(rt_matched, outliers)

    if len(rt_matched) < 3:
        if return_model:
            return rt_arr, None
        else:
            return rt_arr

    if mode == 'linear_interpolation':
        # add zero and rt_max to the beginning and the end
        rt_matched = np.concatenate(([0], rt_matched, [rt_max + rt_matched[-1] - rt_ref[-1]]))
        rt_ref = np.concatenate(([0], rt_ref, [rt_max]))
        f = interp1d(rt_matched, rt_ref, fill_value='extrapolate')
        if return_model:
            return f(rt_arr), f
        else:
            return f(rt_arr)


def rt_anchor_selection(data_list, num=50, noise_tol=0.3, mz_tol=0.01):
    """
    To select anchors for retention time correction. The anchors are commonly detected in the provided
    data, of high intensity, with good peak shape, and equally distributed in the analysis time.

    The number of anchors

    Parameters
    ----------
    data_list: list
        A list of MSData objects or file names of the ouput .txt files.
    num: int
        The number of anchors to be selected.
    noise_tol: float
        The noise level for the anchors. Suggestions: 0.3 or lower.

    Returns
    -------
    anchors: list
        A list of anchors (dict) for retention time correction.
    """

    if isinstance(data_list[0], str):
        # check files and choose the one with the highest total intensity as reference files
        total_int = []
        for file in data_list:
            table = pd.read_csv(file, sep="\t", low_memory=False)
            total_int.append(np.sum(table["peak_height"]))
        ref_idx = np.argmax(total_int)

        table = pd.read_csv(data_list[ref_idx], sep="\t", low_memory=False)
        table = table.sort_values(by="m/z")
        mzs = table["m/z"].values
        n_scores = table["noise_level"].values
        v = [False]
        for i in range(1, len(mzs) - 1):
            if mzs[i] - mzs[i - 1] > mz_tol and mzs[i + 1] - mzs[i] > mz_tol and n_scores[i] < noise_tol:
                v.append(True)
            else:
                v.append(False)
        v.append(False)
        valid_mzs = mzs[v]
        valid_rts = table["RT"].values[v]
        valid_ints = table["peak_height"].values[v]
        valid_mzs = valid_mzs[np.argsort(valid_ints)[-num:]]
        valid_rts = valid_rts[np.argsort(valid_ints)[-num:]]
        # sort the results by retention time
        valid_mzs = valid_mzs[np.argsort(valid_rts)]
        valid_rts = valid_rts[np.argsort(valid_rts)]

        train_idx, _ = _split_to_train_test(valid_rts)

        return valid_mzs[train_idx], valid_rts[train_idx]


def _split_to_train_test(array, interval=0.3):
    """
    To split the selected anchors into training and testing sets.

    Parameters
    ----------
    array: numpy.ndarray
        The retention times of the selected anchors.
    interval: float
        The time interval for splitting the anchors.

    Returns
    -------
    train_idx: list
        The indices of the training set.
    test_idx: list
        The indices of the testing set.
    """

    train_idx = [0]
    test_idx = []
    for i in range(1, len(array)):
        if array[i] - array[train_idx[-1]] < interval:
            test_idx.append(i)
        else:
            train_idx.append(i)

    return train_idx, test_idx


class Feature:
    """
    A class to model a feature in mass spectrometry data. Generally, a feature is defined as
    a unique pair of m/z and retention time.
    """

    def __init__(self, file_number=1):
        """
        Define the attributes of a aligned feature.
        """

        # summarized information
        self.id = None  # index of the feature
        self.reference_file = None  # the reference file
        self.mz = 0.0  # m/z
        self.rt = 0.0  # retention time
        self.highest_intensity = 0.0  # the highest peak height from individual files (which is the reference file)
        self.best_ms2 = None  # the best MS2
        self.gaussian_similarity = 0.0  # Gaussian similarity
        self.noise_level = 0.0  # noise level
        self.asymmetry_factor = 0.0  # asymmetry factor
        self.fill_percentage = 0.0  # fill percentage, in %
        self.charge_state = 1  # charge state
        self.is_isotope = False  # is isotope
        self.isotopes = None  # isotopes
        self.is_in_source_fragment = False  # is in-source fragment
        self.adduct_type = None  # adduct type

        self.annotation = None  # annotation
        self.search_mode = None  # 'identity search', 'hybrid search', or 'mzrt_search'
        self.formula = None  # molecular formula
        self.similarity = None  # similarity score (0-1)
        self.matched_peak_number = None  # number of matched peaks
        self.smiles = None  # SMILES
        self.inchikey = None  # InChIKey
        self.matched_ms2 = None  # matched MS2

        # individual files
        self.mz_seq = np.zeros(file_number)  # m/z values from individual files
        self.rt_seq = np.zeros(file_number)  # retention time values from individual files
        self.peak_height_seq = np.zeros(file_number)  # peak height from individual files
        self.peak_area_seq = np.zeros(file_number)  # peak area from individual files
        self.ms2_seq = []  # best MS2 from individual files
        self.detected_seq = np.zeros(file_number, dtype=bool)  # whether the feature is detected in individual files
        self.roi_id_seq = -np.ones(file_number,
                                   dtype=int)  # ROI ID from individual files (-1 if not detected or gap filled)

        # statistical analysis
        self.fold_change = None  # fold change
        self.t_test_p = None  # t-test p-value
        self.adjusted_t_test_p = None  # adjusted t-test p-value

    def calculate_mzrt(self):
        """
        Calculate the m/z and retention time of the feature by averaging the values.
        """

        self.mz = np.mean(self.mz_seq[self.detected_seq])
        self.rt = np.mean(self.rt_seq[self.detected_seq])


import pandas as pd
import numpy as np


def calculate_fdr_vs_score(search, min_matched_peak=4, score_range=np.arange(0.60, 0.96, 0.05)):
    df = pd.read_csv(f'{search}_results_mces.tsv', sep='\t')

    # Filter the dataframe
    df = df[(pd.notnull(df['mces_dist'])) & (df['ion_mode_match'])].reset_index(drop=True)
    df = df[df['matched_peak'] >= min_matched_peak].reset_index(drop=True)

    # Define the match condition
    df['match'] = df['mces_dist'] <= mces_dist_cutoff

    fdr_results = []
    for score_cutoff in score_range:
        df_filtered = df[df['score'] >= score_cutoff]
        if df_filtered.empty:
            fdr = np.nan
        else:
            fdr = (df_filtered['match'] == False).sum() / df_filtered.shape[0]
        fdr_results.append((score_cutoff, fdr))

    return pd.DataFrame(fdr_results, columns=['score_cutoff', 'fdr'])


def calculate_fdr_vs_score_merged(min_matched_peak=4, score_range=np.arange(0.60, 0.96, 0.05)):
    df1 = pd.read_csv('k0_results_mces.tsv', sep='\t')
    df2 = pd.read_csv('k10_results_mces.tsv', sep='\t')

    df = pd.concat([df1, df2]).reset_index(drop=True)

    # Define the match condition
    df['match'] = df['mces_dist'] <= mces_dist_cutoff

    # Filter the dataframe
    df = df[(pd.notnull(df['mces_dist'])) & (df['ion_mode_match'])].reset_index(drop=True)
    df = df[df['matched_peak'] >= min_matched_peak].reset_index(drop=True)
    # sort by score
    df = df.sort_values('score', ascending=False)

    fdr_results = []
    for score_cutoff in score_range:
        df_filtered = df[df['score'] >= score_cutoff]
        df_filtered = df_filtered.drop_duplicates(subset=['qry_db_id', 'matched_db_id'])
        if df_filtered.empty:
            fdr = np.nan
        else:
            fdr = (df_filtered['match'] == False).sum() / df_filtered.shape[0]
        fdr_results.append((score_cutoff, fdr))

    return pd.DataFrame(fdr_results, columns=['score_cutoff', 'fdr'])


def gen_all_results(min_matched_peak=4):
    fdr_df_k0 = calculate_fdr_vs_score('k0', min_matched_peak)
    fdr_df_k0['mode'] = 'k0'

    fdr_df_k10 = calculate_fdr_vs_score('k10', min_matched_peak)
    fdr_df_k10['mode'] = 'k10'

    fdr_df = calculate_fdr_vs_score_merged(min_matched_peak)
    fdr_df['mode'] = 'merged'

    # merge dfs
    fdr_df = pd.concat([fdr_df_k0, fdr_df_k10, fdr_df]).reset_index(drop=True)

    fdr_df.to_csv(f'fdr_df_minpeak{min_matched_peak}.tsv', sep='\t', index=False)


import matplotlib.pyplot as plt


def plot(min_matched_peak):
    # Read the data
    df = pd.read_csv(f'fdr_df_minpeak{min_matched_peak}.tsv', sep='\t')

    # Set the font to Arial
    plt.rcParams['font.family'] = 'Arial'

    # Create the plot
    plt.figure(figsize=(5, 3))

    # Plot lines and scatter points for each mode
    for i, mode in enumerate(df['mode'].unique()):
        mode_data = df[df['mode'] == mode]
        label = ['Non-scaled', 'Scaled', 'Search both'][i]
        plt.plot(mode_data['score_cutoff'], mode_data['fdr'], label=label)
        plt.scatter(mode_data['score_cutoff'], mode_data['fdr'], s=15)

    # Customize the plot
    plt.xlabel('Score cutoff', fontname='Arial', fontsize=14, labelpad=8, color='0.2')
    plt.ylabel('FDR', fontname='Arial', fontsize=14, labelpad=6, color='0.2')
    plt.legend(prop={'family': 'Arial'})
    plt.grid(True, linestyle='--', alpha=0.3)
    plt.ylim(0, 0.6)

    plt.title(f'Minimum matched peak: {min_matched_peak}', fontname='Arial', fontsize=14, pad=10, color='0.2')

    # Adjust frame color to 0.4 (gray) and increase tick length
    ax = plt.gca()
    for spine in ax.spines.values():
        spine.set_edgecolor('0.4')
    ax.tick_params(width=1, length=2, color='0.4')

    # Ensure all text uses Arial
    for text in plt.gca().get_children():
        if isinstance(text, plt.Text):
            text.set_fontname('Arial')

    # Show the plot
    plt.tight_layout()

    # save as svg
    plt.savefig(f'fdr_plot_minpeak{min_matched_peak}.svg', transparent=True)
    plt.show()


if __name__ == '__main__':
    mces_dist_cutoff = 4

    # gen_all_results(min_matched_peak=3)
    # gen_all_results(min_matched_peak=4)
    # gen_all_results(min_matched_peak=5)
    # gen_all_results(min_matched_peak=6)

    plot(min_matched_peak=3)
    plot(min_matched_peak=4)
    plot(min_matched_peak=5)
    plot(min_matched_peak=6)
